#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Primitive two-dimensional **convex polygon** (i.e., non-self-intersecting
polygon such that all line segments between any two points on the polygon
boundary remain strictly inside the polygon) functionality.
'''

# ....................{ IMPORTS                           }....................
import math
import numpy as np
from betse.exceptions import BetseMathPolygonException
from betse.util.type.types import type_check, SequenceTypes

# ....................{ EXCEPTIONS                        }....................
def die_unless_polygon(*polygons: SequenceTypes) -> None:
    '''
    Raise an exception unless all passed sequences are two-dimensional
    polygons.

    Parameters
    ----------
    polygons : tuple[SequenceTypes]
        Tuple of all sequences to be validated.

    Raises
    ----------
    BetseMathPointException
        If any such sequence is *not* a two-dimensional polygon.

    See Also
    ----------
    :func:`is_polygon`
        Further details.
    '''

    if not is_polygon(*polygons):
        for polygon in polygons:
            if not is_polygon(polygon):
                raise BetseMathPolygonException(
                    'Sequence not a two-dimensional polygon '
                    '(i.e., length < 3 or item length != 2): {!r}'.format(
                        polygon))

# ....................{ TESTERS                           }....................
@type_check
def is_polygon(*polygons: SequenceTypes) -> bool:
    '''
    ``True`` only if all passed sequences are **two-dimensional polygons**
    (i.e., contain at least three two-dimensional points).

    Parameters
    ----------
    polygons : tuple[SequenceTypes]
        Tuple of all sequences to be tested.

    Returns
    ----------
    bool
        ``True`` only if these sequences are all two-dimensional polygons.

    See Also
    ----------
    :func:`geopoint.is_point`
        Further details on two-dimensional points.
    '''

    # Avoid circular import dependencies.
    from betse.util.math.geometry import geopoint

    # Defer to the existing is_point() tester.
    return all(
        len(polygon) >= 3 and geopoint.is_point(*polygon)
        for polygon in polygons)


@type_check
def is_convex(polygon: SequenceTypes) -> bool:
    '''
    ``True`` only if the passed polygon is **strictly convex** (i.e., is
    non-self-intersecting such that all interior angles are strictly
    between zero and a straight angle).

    Parameters
    ----------
    polygon : SequenceTypes
        Two-dimensional sequence of all points defining the polygon to be
        tested such that:

        * The first dimension indexes each such point (in arbitrary order).
        * The second dimension indexes each coordinate of this point such that:

          * The first item is the X coordinate of this point.
          * The second item is the Y coordinate of this point.

    Returns
    ----------
    bool
        ``True`` only if this polygon is strictly convex.

    Raises
    ----------
    BetseMathPointException
        If this sequence is *not* a two-dimensional polygon.

    See Also
    ----------
    https://stackoverflow.com/a/45372025/2809027
        StackOverflow answer strongly inspiring this implementation. The credit
        entirely goes to Rory Daulton for mastering this non-trivial problem.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.numeric.floats import PI, TWO_PI

    # If this sequence is *NOT* a polygon, raise an exception.
    die_unless_polygon(polygon)

    # NOTES:  1.  Algorithm: the signed changes of the direction angles
    #             from one side to the next side must be all positive or
    #             all negative, and their sum must equal plus-or-minus
    #             one full turn (2 pi radians). Also check for too few,
    #             invalid, or repeated points.
    #         2.  No check is explicitly done for zero internal angles
    #             (180 degree direction-change angle) as this is covered
    #             in other ways, including the `n < 3` check.

    # Needed for any bad points or direction changes.
    try:
        # Get starting information.
        old_x, old_y = polygon[-2]
        new_x, new_y = polygon[-1]
        new_direction = math.atan2(new_y - old_y, new_x - old_x)
        angle_sum = 0.0

        # Check each point (the side ending there, its angle) and accumulated
        # angles.
        for ndx, newpoint in enumerate(polygon):
            # Update point coordinates and side directions and check side
            # length.
            old_x, old_y, old_direction = new_x, new_y, new_direction
            new_x, new_y = newpoint
            new_direction = math.atan2(new_y - old_y, new_x - old_x)

            # If repeated consecutive points, non-convex.
            if old_x == new_x and old_y == new_y:
                return False

            # Calculate & check the normalized direction-change angle.
            angle = new_direction - old_direction

            # Make it in half-open interval (-Pi, Pi].
            if angle <= -PI:
                angle += TWO_PI
            elif angle > PI:
                angle -= TWO_PI

            # If first time through loop, initialize orientation.
            if ndx == 0:
                if angle == 0.0:
                    return False

                orientation = 1.0 if angle > 0.0 else -1.0
            # Else if other time through loop, check orientation is stable.
            elif orientation * angle <= 0.0:  # not both pos. or both neg.
                return False

            # Accumulate the direction-change angle.
            angle_sum += angle

        # Check that the total number of full turns is plus-or-minus 1.
        return abs(round(angle_sum / TWO_PI)) == 1

    #FIXME: Clearly non-ideal. Refactor the above implementation to avoid
    #raising spurious exceptions in the first place.
    except (ArithmeticError, TypeError, ValueError):
        return False  # any exception means not a proper convex polygon


#FIXME: For orthogonality, consider refactoring to accept a single "polygon"
#parameter rather than four separate vertex parameters. Luminescent tree slime!
#FIXME: Consider raising an exception unless the four passed vertices are
#oriented counterclockwise. Fearless furballs without peers!
@type_check
def is_cyclic_quad(
    A: SequenceTypes,
    B: SequenceTypes,
    C: SequenceTypes,
    D: SequenceTypes,
) -> bool:
    '''
    ``True`` only if the quadrilateral represented by the passed four vertices
    is **cyclic** (i.e., counterclockwise-oriented).

    Parameters
    ----------
    A, B, C, D : SequenceTypes
        Four vertices of the quadrilateral to be tested, assumed to be oriented
        counterclockwise.

    Returns
    ----------
    bool
        ``True`` only if this quadrilateral is cyclic.
    '''

    # Avoid circular import dependencies.
    from betse.lib.numpy import nparray, npscalar

    # Coerce these sequences to Numpy arrays for efficiency.
    A = nparray.from_iterable(A)
    B = nparray.from_iterable(B)
    C = nparray.from_iterable(C)
    D = nparray.from_iterable(D)

    # Lengths of the four edges constructed from these vertices.
    a = np.linalg.norm(B - A)
    b = np.linalg.norm(C - B)
    c = np.linalg.norm(D - C)
    d = np.linalg.norm(A - D)

    # Calculate the length of the two diagonals.
    e = np.sqrt(((a * c + b * d) * (a * d + b * c)) / (a * b + c * d))
    f = np.sqrt(((a * c + b * d) * (a * b + c * d)) / (a * d + b * c))

    # For a cyclic quad, the product between the two diagonals equals
    # the product between the two adjacent sides.
    lhs = np.round(e*f, 15)
    rhs = np.round(a*c + b*d, 15)

    # Non-standard Numpy-specific boolean encapsulating this truth value.
    test_bool = lhs == rhs

    # Coerce this into a standard Numpy-agnostic boolean for safety.
    return npscalar.to_python(test_bool)

# ....................{ ORIENTERS                         }....................
#FIXME: Unit test us up.
@type_check
def orient_counterclockwise(polygon: SequenceTypes) -> SequenceTypes:
    '''
    Positively reorient the passed polygon, returning a copy of this polygon
    whose vertices are **positively oriented** (i.e., sorted in
    counter-clockwise order).

    Parameters
    ----------
    polygon : SequenceTypes
        Two-dimensional sequence of all points defining the possibly non-convex
        two-dimensional polygon to be positively oriented such that:

        * The first dimension indexes each such point (in arbitrary order).
        * The second dimension indexes each coordinate of this point such that:

          * The first item is the X coordinate of this point.
          * The second item is the Y coordinate of this point.

        Note that this function expects the type of this sequence to define an
        ``__init__()`` method accepting a passed iterable as its first and only
        positional argument. Unsurprisingly, all builtin sequences (e.g.,
        :class:`tuple`, :class:`list`) *and* :mod:`numpy` sequences (e.g.,
        :class:`numpy.ndarray`) satisfy this requirement.

    Returns
    ----------
    SequenceTypes
        Copy of this polygon positively oriented. The type of this sequence is
        the same as that of the original polygon.
    '''

    # Avoid circular import dependencies.
    from betse.lib.numpy import nparray

    # If this sequence is *NOT* a polygon, raise an exception.
    die_unless_polygon(polygon)

    # Numpy array corresponding to this sequence. While polygon reorientation
    # is feasible in pure-Python, the Numpy-based approach is significantly
    # more efficient as the number of polygon edges increases.
    poly_verts = nparray.from_iterable(polygon)

    # Centre point of this polygon,
    poly_centre = poly_verts.mean(axis=0)

    # One-dimensional Numpy array indexing each vertex of this polygon such
    # that each element is the angle in radians between the positive X-axis and
    # that vertex, derived according to the classic mnemonic SOHCAHTOA:
    #
    #              opposite             (opposite)
    #              --------             (--------)
    #     tan(θ) = adjacent  -->  arctan(adjacent) = θ
    #
    # To sort vertices in a counter-clockwise rotation around the centre point
    # of this polygon rather than around the origin (i.e., the point (0, 0)),
    # each vertex is translated from the origin to this centre point *BEFORE*
    # obtaining this vertex's angle. Assuming standard orientation for a right
    # triangle sitting at the centre point of this polygon:
    #
    # * The opposite edge is the Y coordinate of the current vertex translated
    #   from the origin onto the centre point.
    # * The adjacent edge is the X coordinate of the current vertex translated
    #   from the origin onto the centre point.
    #
    # For safety, the np.arctan2() function intended exactly for this purpose
    # rather than the np.arctan() function intended for general-purpose
    # calculation is called. For further details, see:
    #     https://en.wikipedia.org/wiki/Atan2
    poly_angles = np.arctan2(
        poly_verts[:,1] - poly_centre[1],
        poly_verts[:,0] - poly_centre[0])

    # One-dimensional Numpy array indexing the index of each vertex of this
    # polygon, sorted in counter-clockwise order.
    poly_verts_sorted_index = poly_angles.argsort()

    # Two-dimensional Numpy array of the polygon to be returned, containing all
    # vertices sorted in this order.
    poly_verts_sorted = poly_verts[poly_verts_sorted_index]

    # Return a sequence of the same type as the passed polygon.
    return nparray.to_iterable(array=poly_verts_sorted, cls=type(polygon))



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
**Cell cluster profile** (i.e., collection of parameters associated with a
subset of the cells in a cell cluster) functionality.
'''

# ....................{ IMPORTS                           }....................
from abc import ABCMeta
from betse.science.tissue.picker.tispickcls import TissuePickerABC
from betse.util.type.types import type_check

# ....................{ SUPERCLASS                        }....................
class CellsProfileABC(object, metaclass=ABCMeta):
    '''
    Abstract base class of all **cell cluster profile** (i.e., collection of
    parameters associated with a subset of the cells in a cell cluster)
    subclasses.

    Instances of this class assign a subset of all cells matching
    subclass-specific criteria (e.g., explicit indexing, randomized selection,
    spatial location) to the corresponding tissue profile.

    Attributes
    ----------
    name : str
        Unique name of the current profile.
    z_order : int
        1-based order in which this profile will be plotted with respect to all
        other plotted profiles. Specifically, profiles with larger z order will
        be plotted over profiles with smaller z order.
    picker : TissuePickerABC
        Object assigning a subset of the cell population to this profile.
    '''

    # ..................{ INITIALIZERS                      }..................
    @type_check
    def __init__(
        self, name: str, z_order: int, picker: TissuePickerABC) -> None:
        '''
        Initialize this cell cluster region.

        Parameters
        ----------
        name : str
            Unique name of this profile.
        z_order : int
            1-based order in which this profile will be plotted with respect to
            all other plotted profiles. Specifically, profiles with larger z
            order will be plotted over profiles with smaller z order.
        picker : TissuePickerABC
            Object assigning a subset of the cell population to this profile.
        '''

        # Classify all passed parameters.
        self.name = name
        self.z_order = z_order
        self.picker = picker

# ....................{ SUBCLASSES                        }....................
class TissueProfile(CellsProfileABC):
    '''
    **Tissue profile** (i.e., cell cluster region unconditionally assigning all
    cells in this region the same initial parameters).

    Since these parameters are typically derived from real-world observation of
    biological tissue, the spacialization of these parameters collectively
    defines a "tissue" and is hence referred to as a **tissue profile**.

    Attributes
    ----------
    is_gj_insular : bool
        ``True`` only if gap junctions originating at cells in this tissue are
        **insular** (i.e., prevented from connecting to cells in other
        tissues), implying these gap junctions to be strictly intra-tissue.

    Attributes (Membrane Diffusion)
    ----------
    Dm_Na : float
        Sodium (Na+) membrane diffusion constant in m2/s.
    Dm_K : float
        Potassium (K+) membrane diffusion constant in m2/s.
    Dm_Cl : float
        Chloride (Cl-) membrane diffusion constant in m2/s.
    Dm_Ca : float
        Calcium (Ca2+) membrane diffusion constant in m2/s.
    Dm_H : float
        Hydrogen (H+) membrane diffusion constant in m2/s.
    Dm_M : float
        Charge balance anion (M-) membrane diffusion constant in m2/s.
    Dm_P : float
        Protein (P-) membrane diffusion constant in m2/s.
    '''

    # ..................{ INITIALIZERS                      }..................
    @type_check
    def __init__(
        self,
        is_gj_insular: bool,
        Dm_Na: float,
        Dm_K: float,
        Dm_Cl: float,
        Dm_Ca: float,
        Dm_M: float,
        Dm_P: float,
        *args, **kwargs
    ) -> None:
        '''
        Initialize this tissue profile.

        Parameters
        ----------
        is_gj_insular : bool
            ``True`` only if gap junctions originating at cells in this tissue
            are **insular** (i.e., prevented from connecting to cells in other
            tissues), implying these gap junctions to be strictly intra-tissue.
        Dm_Na : float
            Sodium (Na+) membrane diffusion constant in m2/s.
        Dm_K : float
            Potassium (K+) membrane diffusion constant in m2/s.
        Dm_Cl : float
            Chloride (Cl-) membrane diffusion constant in m2/s.
        Dm_Ca : float
            Calcium (Ca2+) membrane diffusion constant in m2/s.
        Dm_H : float
            Hydrogen (H+) membrane diffusion constant in m2/s.
        Dm_M : float
            Charge balance anion (M-) membrane diffusion constant in m2/s.
        Dm_P : float
            Protein (P-) membrane diffusion constant in m2/s.

        All remaining parameters are passed as is to the superclass constructor.
        '''

        # Initialize our superclass with all remaining parameters.
        super().__init__(*args, **kwargs)

        # Classify all passed parameters.
        self.is_gj_insular = is_gj_insular
        self.Dm_Na = Dm_Na
        self.Dm_K  = Dm_K
        self.Dm_Cl = Dm_Cl
        self.Dm_Ca = Dm_Ca
        self.Dm_M  = Dm_M
        self.Dm_P  = Dm_P


#FIXME: Actually do something here.
class CutProfile(CellsProfileABC):
    '''
    **Cut profile** (i.e., cell cluster region to be permanently removed by a
    cutting event triggered during the simulation phase).

    There exists a many-to-one relation between cut profiles and cutting
    events. That is to say, each cutting event references zero or more cut
    profiles. While unwieldy, this disconnection permits cutting event (but
    *not* cut profile) parameters to be modified without requiring simulation
    reseeding or reinitialization. Welcome to the code jungle.
    '''

    pass



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Unit tests for the :mod:`betse.util.math.geometry.geopolyconvex` submodule.
'''

# ....................{ IMPORTS                           }....................

# ....................{ TESTS                             }....................
def test_clip() -> None:
    '''
    Unit test the :func:`betse.util.math.geometry.geopolyconvex.clip`
    function.
    '''

    # Defer heavyweight imports.
    from betse.util.math.geometry.polygon import geopolyconvex

    # Rectangle to clip all subject polygons by, oriented randomly rather than
    # counter-clockwise. Due to trivial limitations of the
    # betse.lib.numpy.nparray.to_iterable() function, this rectangle is defined
    # as a list of lists rather than tuple of tuples.
    clip_rectangle = [[2, 2], [2, -1], [-1, -1], [-1, 2],]

    # Parallelogram to be clipped, oriented randomly rather than
    # counter-clockwise.
    subject_parallelogram = [[0, 1], [2, 3], [3, 3], [1, 1]]

    # Assert a parallelogram clipped by this rectangle to produce a smaller
    # parallelogram of the same general shape. Since the
    # test_clip_counterclockwise() test exercises all other edge cases, this
    # single assertion suffices here.
    assert geopolyconvex.clip(
        subject_polygon=subject_parallelogram,
        clip_polygon=clip_rectangle,
    # Smaller parallelogram clipped by this rectangle, oriented
    # counter-clockwise. Due to issues in the underlying numpy.ndarray.tolist()
    # method internally called by this call, point coordinates are of differing
    # types. While odd, this does *NO* harm and is thus ignorable.
    ) == [[1.0, 2.0], [0, 1], [1, 1], [2.0, 2.0],]


def test_clip_counterclockwise() -> None:
    '''
    Unit test the
    :func:`betse.util.math.geometry.geopolyconvex.clip_counterclockwise`
    function.
    '''

    # Defer heavyweight imports.
    from betse.util.math.geometry.polygon import geopolyconvex

    # Rectangle to clip all subject polygons by, oriented counter-clockwise.
    clip_rectangle = ((2, 2), (-1, 2), (-1, -1), (2, -1))

    # Smaller rectangle inside this rectangle, oriented counter-clockwise.
    small_rectangle = ((1, 1), (0, 1), (0, 0), (1, 0))

    # Assert a parallelogram clipped by this rectangle to produce a smaller
    # parallelogram of the same general shape.
    assert geopolyconvex.clip_counterclockwise(
        # Parallelogram to be clipped, oriented counter-clockwise.
        subject_polygon=((3, 3), (2, 3), (0, 1), (1, 1)),
        clip_polygon=clip_rectangle,
    # Smaller parallelogram clipped by this rectangle, oriented
    # counter-clockwise.
    ) == ((2, 2), (1, 2), (0, 1), (1, 1))

    # Assert a small rectangle inside this rectangle to *NOT* clip this small
    # rectangle and hence produce the same small rectangle.
    assert geopolyconvex.clip_counterclockwise(
        subject_polygon=small_rectangle, clip_polygon=clip_rectangle,
    ) == small_rectangle

    # Assert a small rectangle outside this rectangle to fully clip this small
    # rectangle and hence produce... nothing. Absolutely nothing.
    assert geopolyconvex.clip_counterclockwise(
        # Small rectangle to be clipped, oriented counter-clockwise.
        subject_polygon=((1, 4), (0, 4), (0, 3), (1, 3)),
        clip_polygon=clip_rectangle,
    ) == ()



#!/usr/bin/env python3
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

import numpy as np


class Gap_Junction(object):
    """
    Defines functions controlling gap junction voltage gating characteristics.

    From Harris et al. J of Neurosci.(1983) 3:79-100.
    For Ambystoma Mexicanum (Axolotl!) early embryo gap junction gating.

    """

    def __init__(self, sim, cells, p):

        self.init(sim, cells, p)


    def init(self, sim, cells, p):

        """
        Initialize the gap junction object to its initial conditions

        """

        # parameters from Harris et al 1983
        self.lamb = 0.0013
        self.A1 = 0.077
        self.A2 = 0.14
        self.Ao = 0.217

        # get updated values of the helper-functions:
        # voltage-dependent time-constant for channel transitioning from closed to open (1/ms, V in mV):

        # we use the absolute value of transjunctional voltage due to symmetric channels and convert to mV:
        V1 = 1.0e3*np.abs(sim.vgj)

        # time constant transitioning from closed to open:
        self.alpha = self.lamb*np.exp(-self.A1*(V1 - p.gj_vthresh))

        # voltage-dependent time constant for channel transitioning from open to closed (1/ms, V in mV):
        beta = self.lamb*np.exp(self.A2*(V1 - p.gj_vthresh))

        self.beta = beta/(1+50*beta)

        # set the initial state to the steady-state conductivity ginf:
        self.gjopen = sim.gj_block*((self.alpha/(self.alpha + self.beta)) + p.gj_min)


    def run(self, sim, cells, p):


        # we use the absolute value of transjunctional voltage due to symmetric channels and convert to mV:
        V1 = 1.0e3 * np.abs(sim.vgj)

        # time constant transitioning from closed to open:
        self.alpha = self.lamb * np.exp(-self.A1 * (V1 - p.gj_vthresh))

        # voltage-dependent time constant for channel transitioning from open to closed (1/ms, V in mV):
        beta = self.lamb * np.exp(self.A2 * (V1 - p.gj_vthresh))

        self.beta = beta / (1 + 50 * beta)

        # update GJ conductivity with incremental change of channel state (time is converted to be in ms):
        # sim.gjopen = sim.gjopen + ((1.0 - sim.gjopen)*self.alpha - (sim.gjopen - p.gj_min)*self.beta)*p.dt*1e3
        dt = p.dt*1e3 # transform the time-step to model's milisecond units

        # advance the gap junction state using Implicit Euler:
        sim.gjopen = (sim.gjopen + dt*(self.alpha + self.beta*p.gj_min))/(1 + self.alpha*dt + self.beta*dt)

        # apply any blocking modulation:
        sim.gjopen = sim.gj_block*sim.gjopen

        # print(sim.gjopen.mean())


class Gap_Junction_i(object):
    """
    Defines simple gap junction voltage gating characteristics.

    """

    def __init__(self, sim, cells, p):

        self.init(sim, cells, p)


    def init(self, sim, cells, p):

        """
        Initialize the gap junction object to its initial conditions

        """

        sim.gjopen = p.gj_min +(-1 / (1 + np.exp((sim.vgj*1e3 + p.gj_vthresh) / p.gj_vgrad)) +
                      1 / (1 + np.exp((sim.vgj*1e3 - p.gj_vthresh) / p.gj_vgrad)))*sim.gj_block

    def run(self, sim, cells, p):


        sim.gjopen = p.gj_min +(-1 / (1 + np.exp((sim.vgj*1e3 + p.gj_vthresh) / p.gj_vgrad)) +
                      1 / (1 + np.exp((sim.vgj*1e3 - p.gj_vthresh) / p.gj_vgrad)))*sim.gj_block


















#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level **application metadata singleton** (i.e., application-wide object
synopsizing application metadata via read-only properties) hierarchy.
'''

#FIXME: The current approach is inefficient in the case of BETSE being
#installed as a compressed EGG rather than an uncompressed directory. In the
#former case, the current approach (namely, the call to
#resources.get_pathname() performed below) silently extracts the entirety of
#this egg to a temporary setuptools-specific cache directory. That's bad. To
#circumvent this, we'll need to refactor the codebase to directly require only
#"file"-like objects rather than indirectly requiring the absolute paths of
#data resources that are then opened as "file"-like objects.
#
#Specifically, whenever we require a "file"-like object for a codebase
#resource, we'll need to call the setuptools-specific
#pkg_resources.resource_stream() function rather than attempting to open the
#path given by a global below. Ultimately, *ALL* of the codebase-specific
#globals declared below (e.g., "DATA_DIRNAME") should go away.

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To avoid race conditions during setuptools-based installation, this
# module may import *ONLY* from modules guaranteed to exist at the start of
# installation. This includes all standard Python and application modules but
# *NOT* third-party dependencies, which if currently uninstalled will only be
# installed at some later time in the installation. Likewise, to avoid circular
# import dependencies, the top-level of this module should avoid importing
# application modules where feasible.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

from abc import ABCMeta
from betse.exceptions import BetseGitException
from betse.util.type.decorator.deccls import abstractproperty
from betse.util.type.decorator.decmemo import property_cached
from betse.util.type.types import type_check, ModuleType, StrOrNoneTypes

# ....................{ SUPERCLASSES                       }....................
class AppMetaABC(object, metaclass=ABCMeta):
    '''
    Abstract base class of all **application metadata singleton** (i.e.,
    application-wide object synopsizing application metadata via read-only
    properties) subclasses.

    Pathnames
    ----------
    This superclass details the structure of this application on the local
    filesystem with properties providing the absolute pathnames of
    application-specific files and directories. These pathnames are intended
    for use by both this application and downstream consumers (e.g., BETSEE).
    For portability, these pathnames intentionally support standard and
    non-standard runtime environments -- including :mod:`setuptools`-installed
    script wrappers *and* PyInstaller-frozen executables.

    Caveats
    ----------
    **Neither this superclass nor any subclass of this superclass may be safely
    accessed from the top-level ``setup.py`` script of this or any other
    application.** This application and hence this superclass is *not*
    guaranteed to exist at setuptools-based installation-time for downstream
    consumers (e.g., BETSEE).

    For that same reason, this superclass intentionally avoids refactoring
    metadata constants defined by the top-level application modules (e.g.,
    :mod:`betse.metadata`, :mod:`betse.metadeps`) into object properties (e.g.,
    refactoring the :attr:`betse.metadata.NAME` constant into an object
    :meth:`name` property). Doing so would prevent their reuse from the
    top-level ``setup.py`` scripts defined by downstream consumers, which would
    render these constants effectively useless for their principal use case.

    Attributes
    ----------
    _is_libs_initted : bool
        ``True`` only if the :meth:`init_libs` method has already been called,
        enabling the optimal :meth:`init_libs_if_needed` method to silently
        reduce to a noop when ``True``.
    '''

    # ..................{ INITIALIZERS                       }..................
    def __init__(self, *args, **kwargs) -> None:
        '''
        Initialize both this application metadata singleton and the current
        application (except mandatory third-party dependencies of this
        application) if this is the first creation of such a singleton for this
        application *or* raise an exception otherwise (i.e., if this
        application has already instantiated such a singleton).

        Design
        ----------
        Callers should avoid explicitly calling either the
        :func:`betse.util.app.meta.appmetaone.set_app_meta` setter or
        :meth:`init_sans_libs` method, unless absolutely required for
        inconceivable reasons. Why? Because this method already calls those
        callables. Specifically, to enforce the contractual guarantees that:

        * Only a single such singleton be instantiated for the lifetime of this
          application, this method internally calls the
          :func:`betse.util.app.meta.appmetaone.set_app_meta` setter preserving
          this guarantee.
        * Instantiating this singleton suffices to initialize this application
          (except third-party dependencies of this application), this method
          internally calls the :meth:`init_sans_libs` method preserving
          this guarantee. For maintainability, callers should avoid doing so.

        Parameters
        ----------
        All parameters are passed as is to the :meth:`init_sans_libs` method.

        Raises
        ----------
        BetseMetaAppException
            If this is *not* the first instantiation of an application
            metadata singleton for the active Python interpreter.

        See Also
        ----------
        :meth:`init_sans_libs`
        :meth:`betse.util.app.meta.appmetaone.set_app_meta`
            Further details.
        '''

        # Avoid circular import dependencies.
        from betse.util.app.meta import appmetaone

        # Localize all instance variables.
        #
        # Note the init_libs() method to *NOT* have been called yet.
        self._is_libs_initted = False

        # Globalize this singleton *BEFORE* subsequent logic (e.g., the
        # logconf.init() call performed by the self.init() call), any of
        # which could potentially require this singleton.
        appmetaone.set_app_meta(self)

        # Initialize this application.
        self.init_sans_libs(*args, **kwargs)


    @type_check
    def init_sans_libs(self) -> None:
        '''
        Initialize this application *except* mandatory third-party dependencies
        of this application, which requires external resources (e.g.,
        command-line options, configuration files) to have been parsed.

        Specifically, the default implementation of this method (in order):

        #. Enables Python's standard handler for segmentation faults.
        #. Globalizes the passed application metadata singleton.
        #. Enables this application's default logging configuration.
        #. Validates but does *not* initialize all mandatory third-party
           dependencies of this application, which the :meth:`init_libs` method
           does so subsequently.
        #. Validates core directories and files required at program startup,
           creating all such directories and files that do *not* already exist
           and are reasonably creatable.
        #. Validates the active Python interpreter (e.g., to support
           multithreading).
        #. Validates the underlying operating system (e.g., to *not* be a
           vanilla Windows shell environment ala either CMD.exe or PowerShell).

        To support caller-specific error handling, this function is intended to
        be called immediately *after* this application begins catching
        otherwise uncaught exceptions.

        Caveats
        ----------
        **This method is idempotent.** This method has been explicitly designed
        to be safely recallable. Each subsequent invocation of this method
        following the first simply reinitializes this application. While
        typically useless, idempotency is required by low-level automation to
        guarantee consistency across repeated runs (e.g., tests, scripts).

        Design
        ----------
        This method intentionally avoids initializing mandatory dependencies,
        as doing so would require late-time startup logic assumed by such
        initiazilation to have already been performed -- namely, finalization
        of the logging configuration and hence command-line argument parsing.
        By compare, this method is internally called by the :meth:`__init__`
        method called as the first statement of this application. Since no
        startup logic has been performed yet, initialization of dependencies
        is deferred until significantly later in the startup process.
        '''

        # Avoid circular import dependencies.
        from betse.lib import libs
        from betse.util.io.error import errfault, errwarning
        from betse.util.io.log.conf import logconf
        from betse.util.os import oses
        from betse.util.py import pys

        # Enable Python's standard handler for segmentation faults *BEFORE*
        # performing any further logic, any of which could conceivably trigger
        # a segmentation fault and hence process termination.
        errfault.handle_faults()

        # Enable our default logging configuration *BEFORE* performing any
        # further logic, any of which could conceivably log messages.
        logconf.init()

        # Enable our default warning configuration *AFTER* enabling our default
        # logging configuration that captures warnings but *BEFORE* performing
        # any validation, which could conceivably attempt to emit warnings.
        errwarning.init()

        # Validate mandatory dependencies. Avoid initializing these
        # dependencies now (e.g., by calling init_libs()). Doing so requires
        # finalization of the logging configuration (e.g., by parsing CLI
        # options), which has yet to happen this early in the lifecycle.
        libs.die_unless_runtime_mandatory_all()

        # Validate the active Python interpreter and operating system *AFTER*
        # mandatory dependencies. While the former (mostly) comprises
        # unenforced recommendations, the latter comprises enforced
        # requirements and should thus be validated first.
        oses.init()
        pys.init()

    # ..................{ INITIALIZERS ~ libs                }..................
    @type_check
    def init_libs(
        self, matplotlib_backend_name: StrOrNoneTypes = None) -> None:
        '''
        Initialize all mandatory third-party dependencies of this application
        with sane defaults.

        Specifically, this method (in no particular order):

        * Reconfigures matplotlib with sane defaults specific to the current
          platform and set of all available third-party GUI frameworks.
        * Initializes NumPy.
        * Initializes Pillow.

        Parameters
        ----------
        matplotlib_backend_name : StrOrNoneTypes
            Name of the matplotlib backend to explicitly enable. Defaults to
            ``None``, in which case this method implicitly enables the first
            importable backend known to be both usable and supported by this
            application (in descending order of preference).
        '''

        # Avoid circular import dependencies.
        from betse.lib.matplotlib.matplotlibs import mpl_config
        from betse.lib.numpy import numpys
        from betse.lib.pickle import pickles
        from betse.lib.pil import pils
        from betse.util.io.log import logs

        # Log this initialization. Since initializing heavyweight third-party
        # dependencies (especially matplotlib) consumes non-trivial time, this
        # message is intentionally exposed to all users by default.
        logs.log_info('Loading third-party dependencies...')

        # Initialize these dependencies in arbitrary order.
        mpl_config.init(backend_name=matplotlib_backend_name)
        numpys.init()
        pickles.init()
        pils.init()

        # Note this method to have been called *AFTER* successfully doing so.
        self._is_libs_initted = True


    def init_libs_if_needed(self, *args, **kwargs) -> None:
        '''
        Initialize all mandatory third-party dependencies of the current
        application with sane defaults if these dependencies have yet to be
        initialized (i.e., if the :meth:`init_libs` method has yet to be
        called) *or* silently reduce to a noop otherwise.

        Parameters
        ----------
        All passed parameters are passed as is to the :meth:`init_libs` method.
        '''

        # Avoid circular import dependencies.
        from betse.util.io.log import logs

        # If the init_libs() method has already been called, log this fact
        # *WITHOUT* recalling that method, thus reducing to a noop.
        if self._is_libs_initted:
            logs.log_debug(
                'Ignoring request to reload third-party dependencies...')
        # Else, the init_libs() method has yet to be called. So, do so.
        else:
            self.init_libs(*args, **kwargs)

    # ..................{ DEINITIALIZERS                     }..................
    def deinit(self) -> None:
        '''
        Deinitialize this application metadata singleton and hence the current
        application, which requires this singleton for basic functioning.

        Caveats
        ----------
        **No application logic may be performed after calling this method.**
        This method nullifies this singleton *and* closes open logfile handles,
        both of which are required for basic application logic.

        See Also
        ----------
        :func:`betse.util.app.meta.appmetoone.deinit`
            Higher-level function encapsulating this lower-level method.
        '''

        # Avoid circular import dependencies.
        from betse.util.app.meta import appmetaone
        from betse.util.io.log import logs
        from betse.util.io.log.conf import logconf

        # Log this attempt.
        logs.log_debug('Deinitializing application metadata singleton...')

        # Disable our default logging configuration, ensuring that open logfile
        # handles are closed on application closure.
        logconf.deinit()

        # Deglobalize this singleton *AFTER* prior logic (e.g., the call to
        # logconf.deinit()), any of which could potentially require this
        # singleton.
        appmetaone.unset_app_meta()

    # ..................{ SUBCLASS ~ properties              }..................
    # Subclasses are required to implement the following abstract properties.

    @abstractproperty
    def _module_metadata(self) -> ModuleType:
        '''
        **Application-wide metadata submodule** (i.e., submodule publishing
        general-purpose metadata as global constants synopsizing the current
        application), imported for caller convenience.

        If this application is:

        * BETSE, this is the :mod:`betse.metadata` submodule.
        * BETSEE, this is the :mod:`betsee.guimetadata` submodule.

        See Also
        ----------
        :meth:`module_metadata`
            Concrete public property validating the module returned by this
            abstract private property to expose the expected attributes.
        '''

        pass


    #FIXME: *THIS SHOULD NO LONGER BE CALLED ANYWHERE.* Excise, please.
    @abstractproperty
    def _module_metadeps(self) -> ModuleType:
        '''
        **Application-wide dependency metadata submodule** (i.e., submodule
        publishing lists of version-pinned dependencies as global constants
        synopsizing all requirements of the current application), imported for
        caller convenience.

        If this application is:

        * BETSE, this is the :mod:`betse.metadeps` submodule.
        * BETSEE, this is the :mod:`betsee.guimetadeps` submodule.

        See Also
        ----------
        :meth:`module_metadeps`
            Concrete public property validating the module returned by this
            abstract private property to expose the expected attributes.
        '''

        pass

    # ..................{ PROPERTIES ~ bool                  }..................
    @property_cached
    def is_git_worktree(self) -> bool:
        '''
        ``True`` only if this application has a Git-based **working tree**
        (i.e., top-level directory containing this application's ``.git``
        subdirectory and ``setup.py`` install script), typically due to this
        application having been installed for developer usage.
        '''

        return self.git_worktree_dirname_or_none is not None

    # ..................{ PROPERTIES ~ module                }..................
    @property_cached
    def module_metadata(self) -> ModuleType:
        '''
        **Application-wide metadata submodule** (i.e., submodule publishing
        general-purpose metadata as global constants synopsizing the current
        application), imported for caller convenience.

        If this application is:

        * BETSE, this is the :mod:`betse.metadata` submodule.
        * BETSEE, this is the :mod:`betsee.guimetadata` submodule.

        See Also
        ----------
        :meth:`_module_metadata`
            Abstract private property returning the module validated by this
            concrete public property to expose the expected attributes.
        '''

        # Avoid circular import dependencies.
        from betse.util.type.obj import objtest

        # Application-wide metadata submodule to be validated and returned.
        module_metadata = self._module_metadata

        # Validate this submodule to expose the expected attributes. For
        # sanity, we validate only a proper subset of the complete set of
        # attributes exposed by this submodule.
        objtest.die_unless_has_attr(
            module_metadata,
            'NAME', 'LICENSE',
            'VERSION', 'VERSION_PARTS',
            'SYNOPSIS', 'DESCRIPTION',
        )

        # Return this validated submodule.
        return module_metadata


    @property_cached
    def module_metadeps(self) -> ModuleType:
        '''
        **Application-wide dependency metadata submodule** (i.e., submodule
        publishing lists of version-pinned dependencies as global constants
        synopsizing all requirements of the current application), imported for
        caller convenience.

        If this application is:

        * BETSE, this is the :mod:`betse.metadeps` submodule.
        * BETSEE, this is the :mod:`betsee.guimetadeps` submodule.

        See Also
        ----------
        :meth:`_module_metadeps`
            Abstract private property returning the module validated by this
            concrete public property to expose the expected attributes.
        '''

        # Avoid circular import dependencies.
        from betse.util.app.meta.appmetamod import (
            MERGE_MODULE_METADEPS_DICTS_NAME)
        from betse.util.type.obj import objtest

        # Application-wide dependency metadata submodule to be validated and
        # returned.
        module_metadeps = self._module_metadeps

        # Validate this submodule to expose the expected attributes. For
        # sanity, we validate only a proper subset of the complete set of
        # attributes exposed by this submodule.
        #
        # Note that the existence of the optional "RequirementCommand" class is
        # intentionally *NOT* validated, due to being... optional.
        objtest.die_unless_has_attr(
            module_metadeps, *MERGE_MODULE_METADEPS_DICTS_NAME)

        # Return this validated submodule.
        return module_metadeps

    # ..................{ PROPERTIES ~ package               }..................
    @property_cached
    def package(self) -> ModuleType:
        '''
        **Root package** (i.e., topmost package for this application, typically
        of the same name as this application and installed into a subdirectory
        of the same name in the ``site-packages`` directory specific to the
        active Python interpreter) for this application.
        '''

        # Avoid circular import dependencies.
        from betse.util.py.module import pypackage

        # Introspection for the glorious victory.
        return pypackage.get_object_type_package_root(obj=self)


    @property_cached
    def package_name(self) -> str:
        '''
        Name of this application's root package (e.g., ``betse`` for BETSE).
        '''

        # Avoid circular import dependencies.
        from betse.util.py.module import pymodule

        # By the power of Grayskull...
        return pymodule.get_name_qualified(module=self.package)

    # ..................{ PROPERTIES ~ package : test        }..................
    @property_cached
    def test_package(self) -> ModuleType:
        '''
        **Root test package** (i.e., topmost package for this application's
        ancillary test suite, typically of the same name as this application
        suffixed by ``_test``) for this application.

        Caveats
        ----------
        **This package is typically not installed with this application,** as
        tests are useless (or at least incidental) for most end user purposes.
        Instead, this package is only distributed with tarballs archiving the
        contents of this application's repository at stable releases time. This
        package is *not* guaranteed to exist and, in fact, typically does not.

        Raises
        ----------
        ImportError
            If this package either does not exist *or* does exist but is
            unimportable (e.g., due to package-scoped side effects at
            importation time).
        '''

        # Avoid circular import dependencies.
        from betse.util.py.module import pymodname

        # Introspection for the glorious victory.
        return pymodname.import_module(module_name=self.test_package_name)


    @property_cached
    def test_package_name(self) -> str:
        '''
        Name of the root package of this application's ancillary test suite
        (e.g., ``betse_test`` for BETSE).

        See Also
        ----------
        :meth:`test_package`
            Further details.
        '''

        # When our powers combine!
        return self.package_name + '_test'

    # ..................{ PROPERTIES ~ dir                   }..................
    @property_cached
    def project_dirname(self) -> str:
        '''
        Absolute dirname of this application's **root project directory**
        (i.e., top-level directory containing this application's installable
        ``pyproject.toml`` file or ``setup.py`` script) if found *or* raise an
        exception otherwise (i.e., if this directory is *not* found).

        Equivalently, this is the same as:

        * The root directory archived by release tarballs for this application.
        * The Git-based working tree for this application (i.e., the top-level
          directory containing this application's ``.git`` subdirectory).

        Caveats
        ----------
        **This directory typically does not exist.** This directory is only
        required during installation by non-developers *or* during development
        by developers. Once this application has been installed in a standard
        (i.e., non-editable) fashion by non-developers, this directory is no
        longer required and hence should *not* be assumed to exist.

        Raises
        ----------
        BetseDirException
            If this directory does *not* exist.
        '''

        # Avoid circular import dependencies.
        from betse.util.path import dirs
        from betse.util.py.module import pypackage

        # Absolute dirname of the parent directory of our top-level package.
        package_dirname = pypackage.get_package_project_dirname(self.package)

        # If this directory is not found, fail; else, return this directory.
        return dirs.dir_or_die(package_dirname)


    @property_cached
    def package_dirname(self) -> str:
        '''
        Absolute dirname of this application's root package directory if found
        *or* raise an exception otherwise (i.e., if this directory is *not*
        found).

        This directory typically resides in the ``site-packages`` subdirectory
        of the system-wide standard lib for the active Python interpreter
        (e.g., ``/usr/lib64/python3.6/site-packages/betse``).

        Raises
        ----------
        BetseDirException
            If this directory does *not* exist.
        '''

        # Avoid circular import dependencies.
        from betse.util.path import dirs
        from betse.util.py.module import pymodule

        # Absolute dirname of the directory yielding our top-level package.
        package_dirname = pymodule.get_dirname(self.package)

        # If this directory is not found, fail; else, return this directory.
        return dirs.dir_or_die(package_dirname)


    @property_cached
    def dot_dirname(self) -> str:
        '''
        Absolute dirname of this application's **root dot directory** (i.e.,
        top-level directory containing this application's user-specific files
        and hence residing in the home directory of the current user), silently
        creating this directory if *not* already found.

        This directory contains user-specific files (e.g., logfiles, profile
        files) both read from and written to at application runtime. These are
        typically plaintext files consumable by external users and third-party
        utilities.

        Locations
        ----------
        Denote:

        * ``{package_name}`` the value of the :meth:`package_name` property for
          this application: e.g., ``betse`` for BETSE.
        * ``{username}`` the name of the current user: e.g., ``leycec``.

        Then the dirname returned by this property is:

        * Under Linux, ``~/.{package_name}/``. This property intentionally does
          *not* currently comply with the `XDG Base Directory Specification`_
          (e.g., ``~/.local/share/betse``), which the authors regard as
          `unhelpful if not blatantly harmful <xkcd Standards_>`__.
        * Under OS X, ``~/Library/Application Support/{package_name}``.
        * Under Windows,
          ``C:\\Documents and Settings\\{username}\\Application Data\\{package_name}``.

        .. _XDG Base Directory Specification:
            http://standards.freedesktop.org/basedir-spec/basedir-spec-latest.html
        .. _xkcd Standards:
            https://xkcd.com/927
        '''

        # Avoid circular import dependencies.
        from betse.util.os.brand import macos, posix, windows
        from betse.util.os.shell import shellenv
        from betse.util.path import dirs, pathnames
        from betse.util.path.pathnames import get_home_dirname

        # Absolute dirname of this directory.
        dot_dirname = None

        # If macOS, prefer a macOS-specific directory.
        if macos.is_macos():
            dot_dirname = pathnames.join(
                get_home_dirname(),
                'Library',
                'Application Support',
                self.package_name,
            )
        # If Windows, prefer a Windows-specific directory.
        elif windows.is_windows():
            # Absolute dirname of the parent directory of all Windows-specific
            # dot directories, conditionally obtained via iteration inspection
            # of a cascading series of decreasingly preferred shell environment
            # variables:
            # * '%LOCALAPPDATA%', the user-specific directory containing local
            #   application data *NOT* synced across remote roaming profiles.
            # * '%APPDATA%', the user-specific directory containing local
            #   application data synced across remote roaming profiles. Since
            #   there exists *NO* demonstrable reason to sync BETSE log files
            #   across remote roaming profiles, '%LOCALAPPDATA%' is preferred to
            #   '%APPDATA%'.
            # * A manually constructed user-specific directory containing local
            #   application data intended to resolve to the same dirname as
            #   '%LOCALAPPDATA%'.
            #
            # In theory, the '%LOCALAPPDATA%' variable should *ALWAYS* be
            # defined under Windows; so, the fallbacks should *NEVER* be
            # leveraged. In practice, edge-case environments like the GitHub
            # Actions Windows runner frequently fail to declare these critical
            # environment variables; so, we do. *shrug*
            dot_parent_dirname = (
                shellenv.get_var_or_default('LOCALAPPDATA',
                shellenv.get_var_or_default('APPDATA',
                pathnames.join(
                    get_home_dirname(),
                    'AppData',
                    'Local',
                ),
            )))

            # Absolute dirname of the Windows-specific BETSE dot directory.
            dot_dirname = pathnames.join(dot_parent_dirname, self.package_name)
        # Else...
        else:
            # If this platform is POSIX-incompatible, raise an exception.
            posix.die_unless_posix()

            # Prefer a POSIX-compatible directory.
            dot_dirname = pathnames.join(
                pathnames.get_home_dirname(), f'.{self.package_name}')

        # Create this directory if needed.
        dirs.make_unless_dir(dot_dirname)

        # Return this dirname.
        return dot_dirname

    # ..................{ PROPERTIES ~ dir : data            }..................
    @property_cached
    def data_dirname(self) -> str:
        '''
        Absolute dirname of this application's top-level data directory if
        found *or* raise an exception otherwise (i.e., if this directory is
        *not* found).

        This directory typically contains application-internal resources (e.g.,
        media files) required at application runtime.

        Raises
        ----------
        BetseDirException
            If this directory does *not* exist.
        '''

        # Avoid circular import dependencies.
        from betse.util.app import apppath

        # Return the absolute dirname of this application-relative directory if
        # this directory exists *OR* raise an exception otherwise.
        return apppath.get_dirname(package=self.package, dirname='data')

    # ..................{ PROPERTIES ~ dir : git             }..................
    @property_cached
    def git_worktree_dirname(self) -> str:
        '''
        Absolute dirname of this application's Git-based **working tree**
        (i.e., top-level directory containing this application's ``.git``
        subdirectory and ``setup.py`` install script) if this application was
        installed in a developer manner *or* raise an exception otherwise
        (i.e., if this directory is *not* found).

        Raises
        ----------
        BetseDirException
            if this application was installed in a developer manner but this
            directory does *not* exist, a paradox that should never occur.
        BetseGitException
            if this application was *not* installed in a developer manner.
        '''

        # Avoid circular import dependencies.
        from betse.util.path import dirs

        # Absolute dirname of this application's Git-based working tree if
        # this application was installed for development or "None" otherwise.
        git_worktree_dirname = self.git_worktree_dirname_or_none

        # If this application is *NOT* under development, fail.
        if git_worktree_dirname is None:
            raise BetseGitException(
                'Package "{}" not under Git-based development.'.format(
                    self.package_name))

        # If this directory is not found, fail; else, return this directory.
        return dirs.dir_or_die(git_worktree_dirname)


    @property_cached
    def git_worktree_dirname_or_none(self) -> StrOrNoneTypes:
        '''
        Absolute dirname of this application's **Git-based working tree**
        (i.e., top-level directory containing this application's ``.git``
        subdirectory and ``setup.py`` install script) if this application was
        installed in a developer manner *or* ``None`` otherwise.

        Returns
        ----------
        StrOrNoneTypes
            Either:

            * The absolute dirname of this application's Git working tree if
              this application was installed in a developer manner: e.g., by

              * ``python3 setup.py develop``.
              * ``python3 setup.py symlink``.

            * ``None`` if this application was installed in a non-developer
              manner: e.g., by

              * ``pip3 install``.
              * ``python3 setup.py develop``.
        '''

        # Avoid circular import dependencies.
        from betse.util.path import gits

        # Behold! It is a one-liner.
        return gits.get_package_worktree_dirname_or_none(self.package)

    # ..................{ PROPERTIES ~ dir : test            }..................
    @property_cached
    def test_dirname(self) -> str:
        '''
        Absolute dirname of this application's ancillary test suite if found
        *or* raise an exception otherwise (i.e., if this directory is *not*
        found).

        Raises
        ----------
        BetseDirException
            If this directory does *not* exist.

        See Also
        ----------
        :meth:`test_package`
            Further details.
        '''

        # Avoid circular import dependencies.
        from betse.util.path import dirs
        from betse.util.py.module import pymodule

        # Absolute dirname of the directory yielding our top-level test suite.
        test_package_dirname = pymodule.get_dirname(self.test_package)

        # If this directory is not found, fail; else, return this directory.
        return dirs.dir_or_die(test_package_dirname)


    @property_cached
    def test_data_dirname(self) -> str:
        '''
        Absolute dirname of this application's top-level test suite data
        directory if found *or* raise an exception otherwise (i.e., if this
        directory is *not* found).

        This directory typically contains application-internal resources (e.g.,
        media files) required at application testing time.

        Raises
        ----------
        BetseDirException
            If this directory does *not* exist.
        '''

        # Avoid circular import dependencies.
        from betse.util.path import dirs

        # Return the absolute dirname of this directory if this directory
        # exists *OR* raise an exception otherwise.
        return dirs.join_or_die(self.test_dirname, '_data')

    # ..................{ PROPERTIES ~ file                  }..................
    @property_cached
    def log_default_filename(self) -> str:
        '''
        Absolute filename of this application's **default logfile** (i.e.,
        plaintext user-specific file to which all messages, warnings, errors,
        and exceptions are logged by default).
        '''

        # Avoid circular import dependencies.
        from betse.util.path import pathnames

        # Return the absolute path of this file.
        return pathnames.join(self.dot_dirname, self.package_name + '.log')


    @property_cached
    def profile_default_filename(self) -> str:
        '''
        Absolute filename of this application's **default profile dumpfile**
        (i.e., user-specific binary file to which profiled statistics are saved
        by default).
        '''

        # Avoid circular import dependencies.
        from betse.util.path import pathnames

        # Return the absolute path of this file.
        return pathnames.join(self.dot_dirname, self.package_name + '.prof')



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
POSIX (Portable Operating System Interface)-specific facilities.
'''

# ....................{ IMPORTS                           }....................
import os
from betse.exceptions import BetseOSException
# from betse.util.io.log import logs
from betse.util.type.decorator.decmemo import func_cached

# ....................{ EXCEPTIONS                        }....................
def die_unless_posix() -> None:
    '''
    Raise an exception unless the current platform is POSIX-compatible.

    See Also
    ----------
    :func:`is_posix`
        Further details.
    '''

    # Avoid circular import dependencies.
    from betse.util.os import oses

    # If the current platform is POSIX-incompatible, raise an exception.
    if not is_posix():
        raise BetseOSException(
            '{} not POSIX-compatible.'.format(oses.get_name()))

# ....................{ TESTERS                           }....................
@func_cached
def is_posix() -> bool:
    '''
    ``True`` only if the current platform is **POSIX-compatible** (i.e.,
    complies with the POSIX (Portable Operating System Interface) standard).

    Typically, this implies this system to *not* be vanilla Microsoft Windows
    and hence to be either:

    * A genuinely POSIX-compliant system.
    * A Cygwin-based Windows application (e.g., CLI terminal, GUI application).
    '''

    return os.name == 'posix'


@func_cached
def is_x11() -> bool:
    '''
    ``True`` only if the active Python interpreter is running under the X
    Window System (hereafter X11), implying this process to be headfull and
    hence support both CLIs and GUIs.

    Caveats
    ----------
    This function returning ``True`` does *not* necessarily imply the current
    platform to be a Linux distribution. Unlike most Linux-centric protocols,
    support for the X11 protocol is sufficiently widespread across non-Linux
    POSIX-compatible platforms as to be effectively cross-platform. Unlike the
    Linux-specific :func:`is_mir` and :func:`is_wayland` testers, this
    Linux-agnostic tester detects X11 by generically detecting the X11-specific
    ``${DISPLAY}`` environment variable rather than the Linux-specific
    ``${XDG_SESSION_TYPE}`` environment variable if any.
    '''

    # Avoid circular import dependencies.
    from betse.util.os.shell import shellenv

    # If the current platform is *NOT* POSIX-compatible, return false.
    if not is_posix():
        return False
    # Else, the current platform is POSIX-compatible.

    # Return true only if the active Python interpreter inherited the
    # X11-specific ${DISPLAY} variable from its parent shell environment.
    return shellenv.is_var('DISPLAY')



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level dynamically linked shared library facilities.

Caveats
----------
Shared library-specific logic is, by definition, operation system-specific
logic and hence poor form. Do so *only* where necessary.
'''

# ....................{ IMPORTS                           }....................
from betse.exceptions import BetseLibException, BetseOSException
from betse.util.type.types import type_check, GeneratorType

# ....................{ CONSTANTS                         }....................
KERNEL_NAME_TO_DLL_FILETYPES = {
    # *ALL* Linux distributions.
    'Linux':   {'so',},

    # OS X and iOS. This currently only includes Mach-O shared libraries,
    # required by Apple mandate to have the filetype defined by the "MH_DYLIB"
    # string constant (i.e., "dylib").
    #
    # Macho-O bundles (i.e., loadable modules) are interchangeable with Macho-O
    # shared libraries for some but *NOT* all operations and hence omitted.
    # Unlike Macho-O shared libraries, Macho-O bundles are *NOT* constrained to
    # have a specific filetype. Common Macho-O bundle filetypes include:
    #
    # * "bundle", the filetype promoted by Apple themselves.
    # * "so", the filetype adopted by most cross-platform applications. Unlike
    #   ELF-compatible shared libraries on Linux, however, Macho-O bundles
    #   cannot be dynamically linked against. Instead, the relationship is
    #   reversed; Macho-O bundles dynamically link against plugin-style APIs
    #   exported by applications.
    'Darwin':  {'dylib',},

    # Windows. Technically, other Windows-specific shared library formats exist
    # (e.g., "ocx" for libraries containing ActiveX controls and "drv" for
    # legacy system drivers). As it remains unclear whether anyone still cares
    # about these formats, however, all are currently omitted.
    'Windows': {'dll',},
}
'''
Dictionary mapping from the name of each possible platform returned by the
:func:`betse.util.os.kernels.get_name` function to the set of all undotted
filetypes signifying shared libraries supported by this platform.
'''

# ....................{ EXCEPTIONS                        }....................
def die_unless_dll(pathname: str) -> None:
    '''
    Raise an exception unless the passed path is that of an existing shared
    library supported by the current platform.

    Parameters
    ----------
    pathname : str
        Absolute or relative path of the shared library to inspect.
    '''

    if not is_dll(pathname):
        raise BetseLibException(
            '"{}" not a shared library.'.format(pathname))

# ....................{ TESTERS                           }....................
@type_check
def is_dll(pathname: str) -> bool:
    '''
    ``True`` only if the passed path is that of an existing shared library
    specific to the current platform (e.g., suffixed by ``.so`` on Linux).

    This function returns ``False`` if either:

    * This path is *not* that of an existing file.
    * This pathname has no filetype.
    * This pathname's filetype is *not* that of a shared library supported by
      the current platform.

    Parameters
    ----------
    pathname : str
        Absolute or relative path of the shared library to inspect.

    Returns
    ----------
    bool
        ``True`` only if this path is that of an existing shared library.
    '''

    # Avoid circular import dependencies.
    from betse.util.path import files, pathnames
    from betse.util.os import kernels

    # Filetype of this pathname if any or None otherwise.
    filetype = pathnames.get_filetype_undotted_or_none(pathname)

    # Name of this platform's low-level kernel (e.g., "Linux", "Darwin").
    kernel_name = kernels.get_name()

    # This path is that of an existing shared library if and only if...
    return (
        # This pathname has a filetype.
        filetype is not None and
        # This filetype is that of a shared library for this platform.
        filetype in KERNEL_NAME_TO_DLL_FILETYPES[kernel_name] and
        # This path is that of an existing file. Since testing this requires a
        # filesystem lookup, this is the slowest test and thus deferred.
        files.is_file(pathname)
    )

# ....................{ GENERATORS                        }....................
#FIXME: Add OS X support as well. Since OS X lacks the "ldd" command, doing
#so will require parsing the output of the equivalent OS X-specific "otool"
#command: e.g.,
#
#    # Under OS X.
#    $ otool -L h3dpost.x
#    h3dpost.x:
#       /opt/local/lib/libmpi.0.dylib (compatibility version 1.0.0, current version 1.0.0)
#       /usr/lib/libSystem.B.dylib (compatibility version 1.0.0, current version 111.1.3)

def iter_linked_filenames(filename: str) -> GeneratorType:
    '''
    Generator iteratively yielding the 2-tuple of the basename and absolute
    filename of each shared library dynamically linked to (and hence required
    at runtime by) the shared library with the passed filename.

    Parameters
    ----------
    filename : str
        Absolute or relative filename of the shared library to inspect.

    Yields
    ----------
    (str, str)
        2-tuple ``(linked_lib_basename, linked_lib_pathname``) such that:

        * ``linked_lib_basename`` is the basename of a shared library
          dynamically linked to the shared library with the passed path.
        * ``linked_lib_pathname`` is the absolute pathname of the same library.
    '''

    # Avoid circular import dependencies.
    from betse.util.os import oses
    from betse.util.os.brand import linux
    from betse.util.os.command import cmdrun
    from betse.util.type.text import regexes

    # If this library does *NOT* exist, raise an exception.
    die_unless_dll(filename)

    # If the current platform is Linux...
    if linux.is_linux():
        # String listing all libraries linked to by this library captured from
        # stdout printed by this command. For example, when passed the absolute
        # path of the file defining the "numpy._core.multiarray" C extension,
        # this command prints stdout resembling:
        #
        # multiarray.cpython-34.so:
        # 	linux-vdso.so.1 (0x00007fff049ca000)
        # 	libopenblas_threads.so.0 => /usr/lib64/libopenblas_threads.so.0 (0x00007f9af3ad5000)
        # 	libm.so.6 => /lib64/libm.so.6 (0x00007f9af37d9000)
        # 	libpython3.4.so.1.0 => /usr/lib64/libpython3.4.so.1.0 (0x00007f9af3357000)
        # 	libc.so.6 => /lib64/libc.so.6 (0x00007f9af2fc1000)
        # 	/lib64/ld-linux-x86-64.so.2 (0x000055e2dd7f3000)
        # 	libpthread.so.0 => /lib64/libpthread.so.0 (0x00007f9af2da5000)
        # 	libdl.so.2 => /lib64/libdl.so.2 (0x00007f9af2ba0000)
        # 	libutil.so.1 => /lib64/libutil.so.1 (0x00007f9af299d000)
        ldd_stdout = cmdrun.get_stdout_or_die(command_words=('ldd', filename))

        # For each line containing a "=>"-delimited basename and absolute path
        # pair and hence ignoring both pseudo-libraries that do *NOT* actually
        # exist (e.g., "linux-vdso") or ubiquitous libraries required by the
        # dynamic linking mechanism and hence guaranteed to *ALWAYS* exist
        # (e.g., "ld-linux-x86-64")...
        for line_match in regexes.iter_matches_line(
            text=ldd_stdout,
            regex=r'^\s+(\S+)\s+=>\s+(\S+)\s+\(0x[0-9a-fA-F]+\)$',
        ):
            # Yield the 2-tuple corresponding exactly to the match groups
            # captured by this match.
            yield line_match.groups()

    # Else, library inspection is currently unsupported on this platform.
    raise BetseOSException(
        'Shared library inspection currently unsupported under {}.'.format(
            oses.get_name()))



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **numeric** (i.e., number in the general sense and hence equally
applicable to both integers and floats) facilities.
'''

# ....................{ IMPORTS                           }....................
from betse.exceptions import BetseNumericException
from betse.util.type.types import (
    type_check, IterableTypes, NumericSimpleTypes)

# ....................{ EXCEPTIONS                        }....................
@type_check
def die_unless_positive(
    *numbers: NumericSimpleTypes, label: str = 'Number') -> None:
    '''
    Raise an exception prefixed by the passed label unless all passed objects
    are positive integers or floats.

    Parameters
    ----------
    numbers : tuple
        Tuple of all objects to be validated.
    label : optional[str]
        Human-readable label prefixing exception messages raised by this
        method. Defaults to a general-purpose string.

    Raises
    ----------
    BetseNumericException
        If any passed object is neither a positive integer nor float.
    '''

    # For each passed number...
    for number in numbers:
        # If this number is non-positive...
        if number <= 0:
            # Raise an exception.
            raise BetseNumericException(
                '{} "{}" not positive.'.format(label.capitalize(), number))

# ....................{ SUMMERS                           }....................
@type_check
def sum_by_index(iterable: IterableTypes, subiterable_index: object) -> (
    NumericSimpleTypes):
    '''
    Summation of each **number** (i.e., integer or float) at the passed key or
    index of each subiterable of the passed iterable.

    Each number at the passed key or index of each subiterable of this iterable
    is summed with the ``+`` operator, implicitly calling the ``__sum__()``
    special method of these numbers. The type of the resulting number is the
    widest type of these numbers. Specifically:

    * If any such number is a float, the returned number is also a float.
    * Else, the returned number is an integer.

    For disambiguity, each number is ideally but *not* necessarily of the same
    numeric type: namely, either all integers or all floats.

    Parameters
    ----------
    iterable : IterableTypes
        Iterable of subiterables to be summed.
    subiterable_index : object
        Object with which to index each subiterable of this iterable. The type
        of this object _must_ be a type accepted by the ``__getitem__()``
        special method of each subiterable. Specifically, if each subiterable
        is a:

        * **Mapping** (e.g., :class:`dict`), this object _must_ be hashable.
        * **Sequence** (e.g., :class:`list`, :class:`tuple`), this object
          *must* be either:

          * An integer.
          * A :func:`slice` object.

    Returns
    ----------
    NumericSimpleTypes
        Summation of each number at this key or index of each subiterable of
        this iterable.
    '''

    # Efficiency and simplicity combine here to form MegaFastSimple.
    return sum(subiterable[subiterable_index] for subiterable in iterable)



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Abstract base classes of all timed event classes.
'''

# ....................{ IMPORTS                           }....................
from abc import ABCMeta  #, abstractmethod
from betse.exceptions import BetseSimEventException
from betse.science.phase.phasecls import SimPhase
from betse.util.io.log import logs
from betse.util.type.types import type_check

# ....................{ SUPERCLASSES                      }....................
class SimEventABC(object, metaclass=ABCMeta):
    '''
    Abstract base class of all timed event classes.

    Each instance of this class parameterizes some user-defined event (e.g.,
    cell removal) to be applied at some simulation time step(s).

    Attributes
    ----------
    _is_fired : bool
        ``True`` only if this event's :meth:`fire` method has already been
        called for a previous time step of the appropriate simulation phase.
    '''

    # ..................{ INITIALIZERS                       }..................
    @type_check
    def __init__(self) -> None:
        '''
        Initialize this event.
        '''

        # Classify all passed parameters.
        self._is_fired = False

    # ..................{ PROPERTIES                         }..................
    # Read-only properties, preventing callers from setting these attributes.

    @property
    def is_fired(self) -> bool:
        '''
        ``True`` only if this event's :meth:`fire` method has already been
        called for a previous time step of the appropriate simulation phase.
        '''

        return self._is_fired

    # ..................{ FIRERS                             }..................
    @type_check
    def fire(self, phase: SimPhase, time_step: float) -> None:
        '''
        Apply this event for the passed time step of the passed simulation
        phase.

        This method notes this event to have been applied, ensuring the
        :meth:`is_fired` property to subsequently report ``True``.

        Parameters
        ----------
        phase : SimPhase
            Current simulation phase.
        time_step : float
            Current time step of this phase being simulated.
        '''

        self._is_fired = True

# ....................{ SUBCLASSES                        }....................
class SimEventSpikeABC(SimEventABC):
    '''
    Abstract base class of all **simulation spike event** (i.e., occurring at
    only a single time step rather than over a range of time steps) subclasses.

    Attributes
    ----------
    _time_step : float
        Time step in seconds (s) at which to trigger this action.
    '''

    # ..................{ INITIALIZERS                     }..................
    @type_check
    def __init__(
        self,
        p: 'betse.science.parameters.Parameters',
        time_step: float,
    ) -> None:
        '''
        Initialize this simulation spike event for the passed simulation
        configuration.

        Parameters
        ----------
        p : betse.science.parameters.Parameters
            Current simulation configuration.
        time_step : float
            Time step in seconds (s) at which to trigger this action.

        Raises
        ----------
        BetseSimEventException
            If this time step is invalid (i.e., *not* in the range
            ``[0, p.sim_time_total)``).
        '''

        # Initialize our superclass.
        super().__init__()

        # If this time step is invalid, log a non-fatal warning.
        #
        # While an invalid time step is arguably questionable, this
        # invalidity is *NOT* fatal for most use cases and sublasses.
        # Ergo, logging a non-fatal warning is saner than raising a
        # fatal exception here.
        if not 0.0 <= time_step < p.sim_time_total:
            logs.log_warning(
                'Event time %f invalid '
                '(i.e., not in range [0.0, %f)).',
                time_step, p.sim_time_total)

        # Classify all passed parameters.
        self._time_step = time_step


class SimEventPulseABC(SimEventABC):
    '''
    Abstract base class of all **simulation pulse event** (i.e., occurring over
    a range of time steps rather than at only a single time step) subclasses.

    Attributes
    ----------
    start_time_step : float
        Time step (s) at which to begin triggering this event.
    stop_time_step : float
        Time step (s) at which to cease triggering this event.
    time_step_rate : float
        Slope of the pair of step functions guaranteeing smooth continuity
        between the background function and this event. Each step function is
        the mirror image of the other reflected across the Y axis. These are:

        * A "step" up from the background function to this event, whose
          mid-point is centered at ``start_time``.
        * A "step" down from this event back to the background function, whose
          mid-point is centered at ``stop_time``.

        If the background function is time-dependent, this slope is a **rate**
        (i.e., change over time). For the :class:`SimEventPulseVoltage`
        subclass, for example, this is the rate in voltage per seconds (V/s) at
        which:

        * The background voltage is first increased to the peak voltage.
        * The peak voltage is later decreased to the background voltage.
    '''

    # ..................{ INITIALIZERS                      }..................
    @type_check
    def __init__(
        self,
        p: 'betse.science.parameters.Parameters',
        start_time_step: float,
        stop_time_step: float,
        time_step_rate: float,
    ) -> None:
        '''
        Initialize this simulation pulse event for the passed simulation
        configuration.

        Parameters
        ----------
        p : betse.science.parameters.Parameters
            Current simulation configuration.
        start_time_step : float
            Time step (s) at which to begin triggering this event.
        stop_time_step : float
            Time step (s) at which to cease triggering this event.
        time_step_rate : float
            Slope of the pair of step functions guaranteeing smooth continuity
            between the background function and this event. See the class
            docstring for details.
        '''

        # Initialize our superclass.
        super().__init__()

        # If this start exceeds this stop time step, raise an exception.
        if start_time_step > stop_time_step:
            raise BetseSimEventException(
                'Start event time {} exceeds stop event time {}.'.format(
                    start_time_step, stop_time_step))

        # FIXME -- these might be warnings, but an exception is not feasible as these are somtimes
        # desired states.

        # # If this start or stop time step are invalid, raise an exception.
        # if not 0.0 <= start_time_step < p.sim_time_total:
        #     raise BetseSimEventException(
        #         'Start event time {} invalid (i.e., not in range '
        #         '[0.0, {})).'.format(start_time_step, p.sim_time_total))
        # if not 0.0 <= stop_time_step < p.sim_time_total:
        #     raise BetseSimEventException(
        #         'Stop event time {} invalid (i.e., not in range '
        #         '[0.0, {})).'.format(stop_time_step, p.sim_time_total))

        # Classify all passed parameters.
        self.start_time_step = start_time_step
        self.stop_time_step = stop_time_step
        self.time_step_rate = time_step_rate



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level non-directory filename facilities.

See Also
----------
:mod:`betse.util.io.iofiles`
    Low-level non-directory file content facilities.
'''

# ....................{ IMPORTS                           }....................
import os, shutil
from betse.exceptions import BetseFileException
from betse.util.io.log import logs
from betse.util.type.types import type_check
from os import path as os_path

# ....................{ EXCEPTIONS                        }....................
@type_check
def die_if_file(pathname: str) -> None:
    '''
    Raise an exception if the passed path is an existing non-directory file
    *after* following symbolic links.

    Raises
    ----------
    BetseFileException
        If any passed path is such a file.

    See Also
    ----------
    :func:`is_file`
        Further details.
    '''

    # If this file exists...
    if is_file(pathname):
        # Avoid circular import dependencies.
        from betse.util.path import paths

        # Raise a human-readable exception.
        raise BetseFileException(
            'Path "{}" already an existing {}.'.format(
                pathname, paths.get_type_label(pathname)))


@type_check
def die_unless_file(*pathnames: str) -> None:
    '''
    Raise an exception unless all passed paths are existing non-directory files
    *after* following symbolic links.

    Raises
    ----------
    BetseFileException
        If any passed path is *NOT* such a file.

    See Also
    ----------
    :func:`is_file`
        Further details.
    '''

    # If at least one passed path is *NOT* an existing non-directory file...
    if not is_file(*pathnames):
        # For each such path...
        for pathname in pathnames:
            # If this path *NOT* such a file, raise an exception.
            if not is_file(*pathnames):
                raise BetseFileException(
                    'File "{}" not found or unreadable.'.format(pathname))


@type_check
def join_or_die(*pathnames: str) -> str:
    '''
    Pathname of the file produced by joining (i.e., concatenating) the passed
    pathnames with the directory separator specific to the current platform if
    this file exists *or* raise an exception otherwise (i.e., if this file does
    *not* exist).

    See Also
    -----------
    :func:`betse.util.path.pathnames.join`
    :func:`die_unless_file`
        Further details.
    '''

    # Avoid circular import dependencies.
    from betse.util.path.pathnames import join

    # Filename producing by joining these pathnames.
    filename = join(*pathnames)

    # If this file does *NOT* exist, raise an exception.
    die_unless_file(filename)

    # Return this dirname.
    return filename

# ....................{ EXCEPTIONS ~ writable             }....................
@type_check
def die_unless_exists_writable(pathname: str) -> None:
    '''
    Raise an exception unless the passed path is an existing writable
    non-directory file *after* following symbolic links.

    Raises
    ----------
    BetseFileException
        If this path is *NOT* such a file.

    See Also
    ----------
    :func:`is_writable`
        Further details.
    '''

    # If this file does not exist or does but is unwritable, raise an
    # exception.
    if not is_exists_writable(pathname):
        raise BetseFileException(
            'File "{}" not found or unwritable.'.format(pathname))

# ....................{ TESTERS                           }....................
@type_check
def is_file(*pathnames: str) -> bool:
    '''
    ``True`` only if all passed paths are existing non-directory files *after*
    following symbolic links.

    If any such path does *not* exist, this function silently ignores this path
    rather than raising an exception.

    Versus :func:`path.isfile`
    ----------
    This function intrinsically differs from the standard :func:`path.isfile`
    function. While the latter returns ``True`` only for non-special files and
    hence `False` for all non-directory special files (e.g., device nodes,
    sockets), this function returns ``True`` for *all* non-directory files
    regardless of whether these files are special or not.

    **Why?** Because this function complies with POSIX semantics, whereas
    :func:`path.isfile` does *not*. The specialness of non-directory files is
    usually irrelevant; in general, it only matters whether these files are
    directories or not. For example, the external ``rm`` command removes only
    non-directory files (regardless of specialness) while the external
    ``rmdir`` command removes only empty directories.

    Parameters
    ----------
    pathnames: tuple[str]
        Tuple of all paths to be tested.

    Returns
    ----------
    bool
        ``True`` only if all such paths are existing non-directory files
        *after* following symbolic links.
    '''

    # Avoid circular import dependencies.
    from betse.util.path import dirs, paths

    # Return true only if...
    return any(
        # This path both exists and is *NOT* a directory...
        paths.is_path(pathname) and not dirs.is_dir(pathname)
        # For each such path.
        for pathname in pathnames
    )


@type_check
def is_empty(pathname: str) -> bool:
    '''
    ``True`` only if the passed path is an **empty non-directory file**
    (i.e., zero-byte file) *after* following symbolic links.

    For generality, this function does *not* raise an exception if this path
    does not exist or does but is an existing directory.
    '''

    return is_file(pathname) and get_size(pathname) == 0


@type_check
def is_executable(pathname: str) -> bool:
    '''
    ``True`` only if the passed path is an **executable non-directory file**
    (i.e., file with the execute bit enabled) *after* following symbolic links.

    For generality, this function does *not* raise an exception if this path
    does not exist or does but is an existing directory.
    '''

    return is_file(pathname) and os.access(pathname, os.X_OK)


@type_check
def is_exists_writable(pathname: str) -> bool:
    '''
    ``True`` only if the passed path is an **existing writable non-directory
    file** (i.e., file with the write bit enabled for the current user or user
    group) *after* following symbolic links.

    For generality, this function does *not* raise an exception if this path
    does not exist or does but is an existing directory.

    See Also
    ----------
    :func:`paths.is_writable`
        Similar tester omitting the existence requirement.
    '''

    return is_file(pathname) and os.access(pathname, os.W_OK)

# ....................{ TESTERS ~ symlink                 }....................
@type_check
def is_symlink(pathname: str) -> bool:
    '''
    ``True`` only if the passed path is an existing symbolic link *before*
    following symbolic links.

    This function does *not* raise an exception if this path does not exist.
    '''

    return os_path.islink(pathname)


@type_check
def is_symlink_valid(pathname: str) -> bool:
    '''
    ``True`` only if the passed path is an existing **non-dangling symbolic
    link** (i.e., symbolic link whose target also exists) *before* following
    symbolic links.

    This function does *not* raise an exception if this path does not exist.
    '''

    # Call path.exists() rather than path.lexists(), as the latter returns True
    # for dangling symbolic links.
    #
    # This is why human-readable function names is a good thing, people.
    return is_symlink(pathname) and os_path.exists(pathname)

# ....................{ GETTERS                           }....................
@type_check
def get_size(filename: str) -> int:
    '''
    Size in bytes of the contents of the file with the passed filename.
    '''

    # If this file does *NOT* exist, raise an exception.
    die_unless_file(filename)

    # Return this filesize.
    return os_path.getsize(filename)

# ....................{ COPIERS                           }....................
@type_check
def copy_overwritable(src_filename: str, trg_filename: str) -> None:
    '''
    Copy the passed source file to the passed target file or directory,
    silently overwriting that target file if that file already exists.

    See Also
    ----------
    :func:`copy`
        Lower-level function wrapped by this higher-level function.
    '''

    # Glory be unto the one-liner for it is beneficent in its pale grandeur!
    return copy(
        src_filename=src_filename,
        trg_filename=trg_filename,
        is_overwritable=True,
    )


@type_check
def copy(
    # Mandatory parameters.
    src_filename: str,
    trg_filename: str,

    # Optional parameters.
    is_overwritable: bool = False,
) -> None:
    '''
    Copy the passed source file to the passed target file or directory,
    raising an exception if that target file already exists unless the
    ``is_overwritable`` parameter is explicitly passed as ``True``.

    If the:

    * Source file is a symbolic link, this link (rather than its transitive
      target) will be copied and hence preserved.
    * Target file is a directory, the basename of the source file will be
      appended to this directory -- much like the standard ``cp`` POSIX
      command.

    For safety, the target file is:

    * Copied in a manner maximally preserving *all* existing metadata of the
      source file. This includes owner, group, permissions, times (e.g.,
      access, creation, modification), and extended attributes (if any).
    * *Not* overwritten by default. If this file already exists, an exception
      is raised unless the ``is_overwritable`` parameter is explicitly passed
      as ``True``. For brevity, consider calling the :func:`copy_overwritable`.

    Parameters
    ----------
    src_filename : str
        Absolute or relative path of the source file to be copied from.
    trg_filename : str
        Absolute or relative path of the target file to be copied to.
    is_overwritable : optional[bool]
        If this target file already exists and this boolean is ``True``, this
        file is silently overwritten; else, an exception is raised. Defaults to
        ``False``.

    Raises
    ----------
    BetseFileException
        If either the:

        * Source file does not exist.
        * Target file already exists *and* ``is_overwritable`` is ``False``.

    See Also
    ----------
    :func:`copy_overwritable`
        Higher-level function wrapping this lower-level function by
        unconditionally passing ``is_overwritable`` parameter as ``True``.
    '''

    # Avoid circular import dependencies.
    from betse.util.path import dirs, paths, pathnames

    # Log this copy.
    logs.log_debug('Copying file: %s -> %s', src_filename, trg_filename)

    # If this source file does *NOT* exist, raise an exception.
    die_unless_file(src_filename)

    # If this target file is a directory, append the basename of the passed
    # source file to this directory -- much like the "cp" POSIX command.
    if dirs.is_dir(trg_filename):
        trg_filename = pathnames.join(
            trg_filename, pathnames.get_basename(src_filename))

    # If this target file already exists but is *NOT* overwritable, raise an
    # exception.
    if not is_overwritable:
        paths.die_if_path(trg_filename)

    # Create the directory containing this target file *BEFORE* calling the
    # shutil.copy2() function, which assumes this directory to exist.
    dirs.make_parent_unless_dir(trg_filename)

    # Perform this copy in a manner preserving metadata and symbolic links.
    shutil.copy2(src_filename, trg_filename, follow_symlinks=False)

# ....................{ REMOVERS                          }....................
@type_check
def remove_file(filename: str) -> None:
    '''
    Remove the non-directory file with the passed filename.

    Parameters
    ----------
    filename : str
        Absolute or relative filename of the file to be removed.

    Raises
    ----------
    BetseFileException
        If this file does *not* exist.
    FileNotFoundError
        If this file did exist at the time this function was called but was
        removed immediately before this function called the low-level
        :func:`os.remove` function -- or, in simpler words, if a filesystem
        race condition occurs.
    '''

    # Log this removal.
    logs.log_debug('Removing file: %s', filename)

    # If this file does *NOT* exist, raise an exception.
    die_unless_file(filename)

    # Remove this file. Note that the os.remove() and os.unlink() functions are
    # identical. (That was a tad silly, Guido.)
    os.remove(filename)


@type_check
def remove_file_if_found(filename: str) -> None:
    '''
    Remove the non-directory file with the passed filename if this file exists
    *or* silently reduce to a noop otherwise (i.e., if this file does *not*
    exist).

    For safety, this function removes this file atomically; in particular, this
    file's existence is *not* explicitly tested for.

    Parameters
    ----------
    filename : str
        Absolute or relative filename of the file to be removed.
    '''

    # Log this removal if the subsequent removal attempt is likely to actually
    # remove a file. Due to race conditions with other threads and processes,
    # this file could be removed after this test succeeds but before the
    # removal is performed. Since this is largely ignorable, the worst case is
    # an extraneous log message. *collective_shrug*
    if is_file(filename):
        logs.log_debug('Removing file: %s', filename)

    # Remove this file atomically. To avoid race conditions with other threads
    # and processes, this operation is *NOT* embedded in an explicit test for
    # file existence. Instead, the Pythonic Way is embraced.
    try:
        os.remove(filename)
    # If this file does *NOT* exist, ignore this exception.
    except FileNotFoundError:
        pass



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
**Simulation modelling enumerations** (i.e., :class:`enum.Enum` subclasses
internally required by simulation modelling phases).
'''

# ....................{ IMPORTS                           }....................
from betse.util.type import enums

# ....................{ ENUMS                             }....................
SimPhaseKind = enums.make_enum(
    class_name='SimPhaseKind',
    member_names=('SEED', 'INIT', 'SIM',),
    is_ordered=True,
    doc='''
Ordered enumeration of all supported types of **simulation phases** (i.e.,
consecutive steps for the process of simulating simulation configurations).

Each member of this enumeration is arbitrarily comparable to each other member.
Each member's value is less than that of another member's value if and only if
the former simulation phase is performed _before_ the latter. Specifically,
this enumeration is a total ordering such that:

    >>> SimPhaseKind.SEED < SimPhaseKind.INIT < SimPhaseKind.SIM
    True

Attributes
----------
SEED : enum
    Seed simulation phase, as implemented by the
    :meth:`betse.science.simrunner.SimRunner.seed` method. This phase creates
    the cell cluster and caches this cluster to an output file.
INIT : enum
    Initialization simulation phase, as implemented by the
    :meth:`betse.science.simrunner.SimRunner.init` method. This phase
    initializes the previously created cell cluster from a cached input file
    and caches this initialization to an output file.
SIM : enum
    Proper simulation phase, as implemented by the
    :meth:`betse.science.simrunner.SimRunner.sim` method. This phase simulates
    the previously initialized cell cluster from a cached input file and caches
    this simulation to an output file.
''')


#!/usr/bin/env python3
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.


import matplotlib.pyplot as plt

class PolyPicker(object):
    """
    Allows the user to interactively select cell polygons from a graph.
    Once selected, polygons turn bright blue.
    The process is complete when the figure window is closed.
    Returns a list of indices to cells, which will have properties
    changed (e.g. ion channels added) depending on how the polypicker
    is used in the main script.

    """

    def __init__(self,cells,p,resume_after_picking):

        self.resume_after_picking = resume_after_picking
        self.picked_indices = []
        self.plotPatch(cells,p)
      #  self.fig.canvas.mpl_connect('pick_event', self.get_inds)
        self.cidpick = self.fig.canvas.mpl_connect('pick_event', self)
        self.cidclose = self.fig.canvas.mpl_connect('close_event', self.on_close)

    def plotPatch(self,cells,p):

        self.fig = plt.figure()
        self.ax = plt.subplot(111)

        for i, poly in enumerate(cells.cell_verts):
            plygn = plt.Polygon(p.um*poly,facecolor='b')
            plygn.set_picker(True)
            plygn.set_alpha(0.25)
            plygn.ind = i
            self.ax.add_patch(plygn)
            self.ax.set_xlabel('Spatial distance x [um]')
            self.ax.set_ylabel('Spatial distance y [um]')
            self.ax.set_title('Select Cells with Mouse Click')

        self.ax.axis('equal')
        self.ax.autoscale_view()

        plt.show(block=False)

    def on_pick(self,event):

        patch = event.artist
        patch.set_alpha(1.0)
        self.index = patch.ind
        self.fig.canvas.draw()
        self.picked_indices.append(self.index)

    def on_close(self, event):
        self.fig.canvas.mpl_disconnect(self.cidclose)
        self.fig.canvas.mpl_disconnect(self.cidpick)
        self.resume_after_picking(self)

    def __call__(self, event):
            self.on_pick(event)






#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level support facilities for Pillow, the Python Image Library
(PIL)-compatible fork implementing most image I/O performed by this
application.
'''

#FIXME: Revisit imageio when the following feature request is resolved in full:
#    https://github.com/imageio/imageio/issues/289

# ....................{ IMPORTS                           }....................
from PIL import Image
from betse.util.io.log import logs
from betse.util.path import pathnames
from betse.util.type.decorator.decmemo import func_cached
from betse.util.type.types import SetType  # type_check,

# ....................{ GETTERS                           }....................
@func_cached
def get_filetypes() -> SetType:
    '''
    Set of all image filetypes supported by the current version of Pillow.

    For generality, these filetypes are *not* prefixed by a ``.`` delimiter.

    Examples
    ----------
        >>> from betse.lib.pil import pils
        >>> pils.get_filetypes()
        {'flc', 'bmp', 'ppm', 'webp', 'j2k', 'jpf', 'jpe', 'pcd'}
    '''

    # Initialize Pillow if uninitialized.
    #
    # If Pillow is uninitialized, the "Image.EXTENSION" dictionary is empty.
    # Since the betse.util.app.meta.appmetaabc.AppMetaABC.init_libs() function
    # already initializes Pillow, explicitly doing so here should typically
    # *NOT* be necessary. Since this getter could technically be called from
    # global scope prior to the initialization performed by "betse.ignition"
    # *AND* since this initialization efficiently reduces to a noop if
    # unnecessary, manually initializing Pillow here is cost-free... and
    # cost-free is the way to be.
    init()

    # Return a set of...
    return set(
        # This filetype stripped of this prefixing "."...
        pathnames.undot_filetype(filetype_dotted)
        # For each "."-prefixed filetype supported by Pillow.
        for filetype_dotted in Image.EXTENSION.keys()
    )

# ....................{ ENUMERATIONS                      }....................
def init() -> None:
    '''
    Initialize Pillow if uninitialized *or* reduce to a noop otherwise (i.e.,
    if Pillow is already initialized).
    '''

    # Log this initialization.
    logs.log_debug('Initializing Pillow...')

    # If Pillow is already initialized, this function internally reduces to an
    # efficient noop in the expected manner (e.g., by accessing a private
    # boolean global).
    Image.init()



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level **application metadata singleton** (i.e., application-wide object
synopsizing application metadata via read-only properties).
'''

# ....................{ TODO                               }....................
#FIXME: Raise an exception if running with superuser privelages. To avoid
#inevitable security issues, Under *NO* circumstances should either BETSE or
#BETSEE ever be run with elevated permissions. To do so, see the following
#canonical answer:
#    https://stackoverflow.com/a/1026626/2809027

#FIXME: Print memory-related metadata when running "betse info" *AND* log
#non-fatal warnings when BETSE is run under a low-memory environment (e.g.,
#< 4GB available free memory). To do so, note the following canonical API:
#    psutil.Process(os.getpid()).get_memory_info()

#FIXME: The "~/.betse" directory grows fairly large fairly quickly. It'd be
#great to emit non-fatal warnings if its size exceeds some reasonable threshold
#(e.g., 1MB).

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To avoid race conditions during setuptools-based installation, this
# module may import *ONLY* from modules guaranteed to exist at the start of
# installation. This includes all standard Python and application modules but
# *NOT* third-party dependencies, which if currently uninstalled will only be
# installed at some later time in the installation. Likewise, to avoid circular
# import dependencies, the top-level of this module should avoid importing
# application modules where feasible.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

from betse.exceptions import BetseMetaAppException
# from betse.util.io.log import logs
from betse.util.type.types import type_check
# from sys import modules

# ....................{ GLOBALS                            }....................
_app_meta = None
'''
**Application metadata singleton** (i.e., application-wide object synopsizing
application metadata via read-only properties).

Caveats
----------
For safety, callers are advised to call the :func:`get_app_meta` getter safely
returning this private singleton rather than directly accessing this private
singleton unsafely.
'''

# ....................{ EXCEPTIONS                         }....................
def die_if_app_meta() -> None:
    '''
    Raise an exception if the application metadata singleton already exists
    (e.g., due to a prior call to the :func:`set_app_meta` function).

    Raises
    ----------
    BetseMetaAppException
        If this singleton already exists.

    See Also
    ----------
    :func:`is_app_meta`
        Further details.
    '''

    # If an application metadata singleton exists, raise an exception.
    if is_app_meta():
        raise BetseMetaAppException(
            'Application metadata singleton already defined '
            '(i.e., appmetaone.set_app_meta() already called).')


def die_unless_app_meta() -> None:
    '''
    Raise an exception unless an application metadata singleton exists (e.g.,
    due to a prior call to the :func:`set_app_meta` function).

    Equivalently, this function raises an exception if this singleton does
    *not* exist.

    Raises
    ----------
    BetseMetaAppException
        If this singleton does *not* exist.

    See Also
    ----------
    :func:`is_app_meta`
        Further details.
    '''

    # If no application metadata singleton exists, raise an exception.
    if not is_app_meta():
        raise BetseMetaAppException(
            'Application metadata singleton undefined '
            '(i.e., appmetaone.set_app_meta() not called).')

# ....................{ TESTERS                            }....................
def is_app_meta() -> bool:
    '''
    ``True`` only if an application metadata singleton exists (e.g., from a
    prior call to the :func:`set_app_meta` function).
    '''

    return _app_meta is not None

# ....................{ GETTERS                            }....................
# Avoid circular import dependencies.
def get_app_meta() -> 'betse.util.app.meta.appmetaabc.AppMetaABC':
    '''
    **Application metadata singleton** (i.e., application-wide object
    synopsizing application metadata via read-only properties) if this
    singleton has already been set by a prior call to the :func:`set_app_meta`
    function *or* raise an exception otherwise (i.eg., if that function has yet
    to be called).

    Returns
    ----------
    AppMetaABC
        Application metadata singleton defined by the most recent call to the
        :func:`set_app_meta` function.

    Raises
    ----------
    BetseMetaAppException
        If the :func:`set_app_meta` function has yet to be called.
    '''

    # If no application metadata singleton has been set, raise an exception.
    die_unless_app_meta()
    # Else, an application metadata singleton has been set.

    # Return this sisgleton.
    return _app_meta

# ....................{ SETTERS                            }....................
@type_check
def set_app_meta(
    # Avoid circular import dependencies.
    app_meta: 'betse.util.app.meta.appmetaabc.AppMetaABC') -> None:
    '''
    Set the **application metadata singleton** (i.e., application-wide object
    synopsizing application metadata via read-only properties) if this is the
    first call to this function *or* raise an exception otherwise (i.e., if
    this function has already been called).

    Caveats
    ----------
    **This function intentionally performs no logging.** Doing so would be
    unproductive. The first call to this function is implicitly performed by
    the :func:`betse.util.app.meta.appmetaabc.AppMetaABC.__init__` method
    *before* logging has been configured. All logging performed by that call
    (but *not* subsequent calls) would be silently squelched, which any sane
    caller would interpret to be a bug.

    Parameters
    ----------
    app_meta : AppMetaABC
        Application metadata singleton to be set.

    Raises
    ----------
    BetseMetaAppException
        If this singleton has already been set.

    See Also
    ----------
    :meth:`betse.util.app.meta.appmetaabc.AppMetaABC.__init__`
        Higher-level method encapsulating this lower-level function.
    '''

    # Avoid circular import dependencies.
    # from betse.util.test.tsttest import is_testing

    # Enable this singleton global to be overwritten be the passed parameter.
    global _app_meta
    # raise ValueError('ugh')
    # print(f'Setting app meta to: {repr(app_meta)}')

    #FIXME: *NO IDEA, BRO.* For some reason, this seemingly sensible check is
    #failing during "tox"-based and *ONLY* "tox"-based testing. Why? No idea.
    #But we gotta have "tox"-based testing. So, let's just quietly ignore this
    #for the moment. It is what it is. *sigh*
    # # If the BETSE test suite is *NOT* currently running tests...
    # # if not is_testing():
    # if 'pytest' not in modules:
    #     # If this singleton has already been set, raise an exception.
    #     die_if_app_meta()
    #     # Else, this singleton has yet to be set.
    # # Else, the BETSE test suite is currently running tests. In this case, avoid
    # # validating whether this singleton has already been set. See the horrifying
    # # "#FIXME:" comment above for further commentary.

    # Set this singleton global to this caller-specific singleton.
    _app_meta = app_meta

# ....................{ SETTERS ~ unset                    }....................
#FIXME: Preserve but refactor as follows:
#
#* Define a new set_app_meta_if_unset() function in this submodule with the
#  following signature, where "cls" is the "AppMetaABC" subclass to be
#  conditionally instantiated by this function if needed:
#    def set_app_meta_if_unset(cls: ClassType, *args, **kwargs) -> (
#       'betse.util.app.meta.appmetaabc.AppMetaABC')
#* Refactor set_app_meta_betse_if_unset() in terms of set_app_meta_if_unset().
def set_app_meta_betse_if_unset(*args, **kwargs) -> (
    # Avoid circular import dependencies.
    'betse.util.app.meta.appmetaabc.AppMetaABC'):
    '''
    Set the application metadata singleton to a BETSE-specific singleton (i.e.,
    instance of the :class:`betse.appmeta.BetseAppMeta` class) if the
    :func:`set_app_meta` function has yet to be called *or* reduce to a noop
    otherwise (i.e., if that function has already been called).

    This is a convenience function simplifying BETSE initialization for
    low-level edge-case automation (e.g.,
    :mod:`betse.science.__init__`, tests). This function is specific to BETSE
    and hence inappropriate for general-purpose application initialization, as
    required by downstream consumers (e.g., the BETSEE GUI).

    Caveats
    ----------
    **This function does not initialize mandatory third-party dependencies.**
    To enable callers to configure such initialization, callers are required to
    explicitly call the
    :meth:`betse.util.app.meta.appmetaabc.AppMetaABC.init_libs` method on the
    object returned by this function.

    Parameters
    ----------
    All parameters are passed as is to the
    :meth:`betse.util.app.meta.appmetaabc.AppMetaABC.init_sans_libs` method.

    Returns
    ----------
    betse.util.app.meta.appmetaabc.AppMetaABC
        Application metadata singleton.
    '''

    # Avoid circular import dependencies.
    from betse.appmeta import BetseAppMeta

    # If no application metadata singleton has been instantiated, do so. Note
    # that doing so implicitly calls the appmetaone.set_app_meta() function on
    # our behalf, which is certainly nice.
    if not is_app_meta():
        BetseAppMeta(*args, **kwargs)
    # An application metadata singleton has now been instantiated.

    # Return this singleton.
    return get_app_meta()

# ....................{ UNSETTERS                          }....................
def unset_app_meta() -> None:
    '''
    Unset the **application metadata singleton** (i.e., application-wide object
    synopsizing application metadata via read-only properties) if such a
    singleton exists *or* raise an exception otherwise (i.e., if no such
    singleton exists).

    Equivalently, this function resets this singleton to its initial state
    (i.e., ``None``).

    Raises
    ----------
    BetseMetaAppException
        If this singleton has not yet been set.

    See Also
    ----------
    :func:`deinit`
        Higher-level function encapsulating this lower-level function.
    '''

    # Enable this singleton global to be overwritten be the passed parameter.
    global _app_meta

    # If this singleton has not yet been set, raise an exception.
    die_unless_app_meta()

    # Revert this singleton global to its initial state.
    #
    # Note that this reversion is intentionally *NOT* logged. This method is
    # typically only called as the last operation at application shutdown, at
    # which time the logging configuration has already been deinitialized.
    # While feasible, logging here would do so under Python's default logging
    # configuration in a non-human-readable format upsetting end users: e.g.,
    #     DEBUG:betsee:Unsetting application metadata singleton...
    _app_meta = None

# ....................{ DEINITIALIZERS                     }....................
def deinit() -> None:
    '''
    Deinitialize the **application metadata singleton** (i.e., application-wide
    object synopsizing application metadata via read-only properties) if such a
    singleton exists *or* silently reduce to a noop otherwise (i.e., if no such
    singleton exists).

    This function effectively deinitializes the current application, which
    requires this singleton for rudimentary functionality throughout the
    codebase.

    Caveats
    ----------
    **No application logic may be safely performed after calling this method.**
    This method nullifies this singleton *and* closes open logfile handles,
    both of which are required for basic application logic.
    '''

    # If an application metadata singleton exists, deinitialize this singleton.
    if is_app_meta():
        _app_meta.deinit()
    # Else, no such singleton exists. In this case, *NO* logic (including
    # logging) may be safely performed. Ergo, noop.



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level :mod:`pytest`-specific abstract base class (ABC) hierarchies.
'''

# ....................{ IMPORTS                            }....................
from abc import ABCMeta

# ....................{ CLASSES                            }....................
# FIXME: The "_first_failure_method_name" attribute documented below is
# dynamically set on the py.test object encapsulating this class instance
# rather than this actual class instance. This may or may not be a bad thing.
class SerialTestABC(metaclass=ABCMeta):
    '''
    Abstract base class running all test methods defined by this concrete
    subclass **serially** (i.e., in the order in which this subclass declares
    these test such that subsequently declared tests depend on the success of
    all previously declared tests).

    On the first failure of such a test, all subsequent such tests will be
    automaticaly marked as xfailing (i.e., failing *without* being run).

    The majority of the black magic required by this class is implemented as
    low-level py.test hooks in the top-level :mod:`betse_func.conftest` plugin.

    Attributes
    ----------
    _first_failure_method_name : str
        Unqualified name of the first failing test method (i.e., method whose
        execution raised an exception) declared by this subclass for the
        current test session if any such method failed *or* ``None`` otherwise
        (i.e., if no such methods have yet to fail).
    '''

    # ..................{ INITIALIZERS                       }..................
    @staticmethod
    def is_test_serial(item) -> bool:
        '''
        ``True`` only if the passed test callable is **serial** (i.e., a method
        of a subclass of this class).

        Serial methods are intended to be run in test method declaration order,
        such that subsequently declared test methods depend on the success of
        all previously declared test methods.

        Parameters
        ----------
        item : pytest.main.Item
            Metadata encapsulating this test callable (e.g., function, method).

        Returns
        -------
        bool
            ``True`` only if this test is serial.
        '''

        # Class of this test callable if this callable is a method *OR* "None".
        test_class = item.parent.cls

        # This callable is intended to be run serially only if the test
        # subclass to which this callable is bound subclasses this abstract
        # base class.
        return test_class is not None and issubclass(test_class, SerialTestABC)



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
CLI-specific functional tests exercising **solver-agnostic simulations** (i.e.,
simulations arbitrarily supporting all simulation solvers, including both the
complete BETSE solver *and* the "fast" equivalent circuit solver).
'''

# ....................{ IMPORTS                           }....................
import pytest
from betse.util.test.pytest.mark.pytfail import xfail
from betse.util.test.pytest.mark.pytskip import (
    skip_unless_matplotlib_anim_writer,)

# ....................{ TESTS                             }....................
def test_cli_sim_compat(betse_cli_sim_compat: 'CLISimTester') -> None:
    '''
    Functional test exercising all simulation subcommands required to validate
    backward compatibility with a temporary simulation configuration file
    (complete with a pickled seed, initialization, and simulation) produced by
    the oldest version of this application for which the current version of
    this application guarantees backward compatibility.

    Design
    ----------
    Validating backward compatibility requires validating that the current
    version of this application can successfully load all:

    * Simulation configuration files loadable by this older version.
    * Pickled seeds, initializations, and simulations saved by this older
      version.

    Technically, running the:

    * ``plot sim`` subcommand would test the loadability of pickled
      simulations.
    * ``sim`` subcommand would test the loadability of pickled initializations.
    * ``init`` subcommand would test the loadability of pickled seeds.

    That said, running the ``init`` and ``sim`` subcommands would erroneously
    overwrite the pickled seeds and initializations saved by this older
    version. Doing so would invite edge-case issues that are best avoided; in
    particular, care would need to be taken to run the ``sim`` subcommand
    *after* the ``init`` subcommand. To ameliorate these concerns, plotting
    subcommands guaranteed to both test the loadability of pickled objects
    *and* be side effect-free are run instead.

    Parameters
    ----------
    betse_cli_sim_compat : CLISimTester
        Object running BETSE CLI simulation subcommands against a temporary
        simulation configuration produced by this older application version.
    '''

    # Test all simulation-specific plotting subcommands on this configuration.
    betse_cli_sim_compat.run_subcommands(
        #FIXME: Replace the following line with the following following line
        #*AFTER* bundling a working v0.5.0-era pickled initialization and
        #simulation with our test suite.
        *betse_cli_sim_compat.SUBCOMMANDS_TRY,
        # *betse_cli_sim_compat.SUBCOMMANDS_PLOT,
    )

    #FIXME: We unavoidably break backward compatibility with respect to pickled
    #objects and hence currently only test backward compatibility with respect
    #to the simulation configuration. This is non-ideal, but sufficient for the
    #moment. Ideally, this should be reverted on releasing a new version.
    # betse_cli_sim_compat.run_subcommands(
    #     ('seed',),
    #     is_overwriting_config=False,)


def test_cli_sim_default(betse_cli_sim_default: 'CLISimTester') -> None:
    '''
    Functional test exercising the default simulation configuration on all
    simulation phases (i.e., seed, initialization, and simulation), principally
    for computational instability.

    Unlike the minified simulation configuration leveraged by all other tests,
    the default simulation configuration leveraged by this test is unmodified
    (except for disabling interactive features, which non-interactive testing
    unavoidably requires). In particular, this configuration is *not* minified
    for efficient usage by tests and thus incurs a significant performance
    penalty. This test exercises only the minimal subset of this configuration
    required to detect computational instabilities in this configuration --
    namely, the first two simulation phases.

    Parameters
    ----------
    betse_cli_sim_default : CLISimTester
        Object running non-minified BETSE CLI simulation subcommands.
    '''

    # Test only the first two simulation-specific subcommands with this
    # configuration, as documented above.
    betse_cli_sim_default.run_subcommands_sim()


# Sadly, all existing higher-level parametrization decorators defined by the
# "betse.util.test.pytest.mark.params" submodule fail to support embedded py.test
# "skipif" and "xfail" markers. Consequently, we leverage the lower-level
# parametrization decorator shipped with py.test itself.
@pytest.mark.parametrize(
    ('writer_name', 'filetype'), (
        pytest.param(
            'avconv', 'mp4',
            marks=skip_unless_matplotlib_anim_writer('avconv')),
        pytest.param(
            'ffmpeg', 'mkv',
            marks=skip_unless_matplotlib_anim_writer('ffmpeg')),

        #FIXME: Research this deeper, please. Are all Mencoder-based writers
        #genuinely broken (doubtful), is this our fault (very possible), or is
        #this a platform- or version-specific issue and hence mostly not our
        #fault (also very possible)?
        # skip_unless_matplotlib_anim_writer('mencoder')(('mencoder', 'avi')),
        pytest.param(
            'mencoder', 'avi', marks=xfail(
                reason='Mencoder-based writers fail with obscure errors.')),

        # ImageMagick only supports encoding animated GIFs and hence is
        # effectively a joke writer. Since it remains supported, however, we
        # test it with a reasonable facsimile of a straight face.
        pytest.param(
            'imagemagick', 'gif',
            marks=skip_unless_matplotlib_anim_writer('imagemagick')),
    ),
)
def test_cli_sim_video(
    betse_cli_sim: 'CLISimTester', writer_name: str, filetype: str) -> None:
    '''
    Functional test simulating at least one animation (and all simulation
    features required by that animation) and encoding that animation as
    compressed video of the parametrized filetype via the matplotlib animation
    writer of the parametrized name.

    Since video encoding is justifiably expensive in both space and time, this
    test encodes only a single animation (rather than all available animations)
    for a single simulation-specific BETSE CLI subcommand (rather than all
    available subcommands). The existing :func:`test_cli_sim_visuals` test
    already exercises all animations for all subcommands, albeit with video
    encoding disabled.

    Parameters
    ----------
    betse_cli_sim : CLISimTester
        Object running BETSE CLI simulation subcommands.
    writer_name : str
        Name of the matplotlib animation writer with which to encode video
        (e.g., `ffmpeg`, `imagemagick`).
    filetype : str
        Filetype of videos to encode with this writer (e.g., `mkv`, `mp4`).
    '''

    # Test the minimum number of simulation-specific subcommands required to
    # exercise video encoding with this configuration. Since the "init"
    # subcommand requiring the "seed" subcommand satisfies this constraint, all
    # subsequent subcommands (e.g., "sim", "plot init") are omitted.
    betse_cli_sim.run_subcommands(('seed',), ('init',),)



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
YAML-backed simulation subconfigurations for exporting plots.
'''

# ....................{ IMPORTS                           }....................
from betse.lib.yaml.yamlalias import yaml_alias, yaml_alias_int_positive
from betse.lib.yaml.abc.yamlabc import YamlABC
from betse.lib.yaml.abc.yamllistabc import YamlList, YamlListItemABC
from betse.science.config.export.confexpabc import SimConfExportABC
from betse.science.config.export.visual.confexpvisabc import (
    SimConfVisualCellsYAMLMixin)
from betse.util.type.types import type_check

# ....................{ SUBCLASSES                        }....................
class SimConfExportPlots(YamlABC):
    '''
    YAML-backed subconfiguration for exporting *all* plots (both in- and
    post-simulation) enabled by the current YAML-formatted simulation
    configuration file.

    This subconfiguration saves (i.e., writes, serializes) in-memory plots to
    on-disk cache, image, and/or video files configured by this configuration.

    Attributes (After Solving)
    ----------
    is_after_sim : bool
        ``True`` only if post-simulation plots are to be displayed and/or saved.
    is_after_sim_show : bool
        ``True`` only if post-simulation plots are to be displayed.
    is_after_sim_save : bool
        ``True`` only if post-simulation plots are to be saved.

    Attributes (After Solving: Single-cell)
    ----------
    plots_cell_after_sim : YamlList
        YAML-backed list of all post-simulation single-cell plots to be
        animated. Ignored if :attr:``is_after_sim`` is ``False``.
    plots_cells_after_sim : YamlList
        YAML-backed list of all post-simulation cell cluster plots to be
        animated. Ignored if :attr:``is_after_sim`` is ``False``.

    Attributes (Image)
    ----------
    image_filetype : str
        Filetype of all image files saved by this configuration. Ignored if
        :attr:`is_after_sim_save` is ``False``.
    image_dpi : int
        Dots per inch (DPI) of all image files saved by this configuration.
        Ignored if :attr:`is_after_sim_save` is ``False``.
    '''

    # ..................{ ALIASES ~ after                   }..................
    is_after_sim_save = yaml_alias(
        "['results options']['after solving']['plots']['save']", bool)
    is_after_sim_show = yaml_alias(
        "['results options']['after solving']['plots']['show']", bool)

    # ..................{ ALIASES ~ save                    }..................
    image_filetype = yaml_alias(
        "['results options']['save']['plots']['filetype']", str)
    image_dpi = yaml_alias_int_positive(
        "['results options']['save']['plots']['dpi']")

    # ..................{ INITIALIZERS                      }..................
    def __init__(self, *args, **kwargs) -> None:

        # Initialize our superclass with all passed parameters.
        super().__init__(*args, **kwargs)

        # Encapsulate low-level lists of dictionaries with high-level wrappers.
        self.plots_cell_after_sim  = SimConfExportPlotCell .make_list()
        self.plots_cells_after_sim = SimConfExportPlotCells.make_list()

    # ..................{ LOADERS                           }..................
    def load(self, *args, **kwargs) -> None:

        # Load our superclass with all passed arguments.
        super().load(*args, **kwargs)

        # Simulation subconfigurations localized for convenience.
        plot_pipelines = self._conf[
            'results options']['after solving']['plots']

        # Load all subconfigurations of this configuration.
        self.plots_cell_after_sim.load(
            conf=plot_pipelines['single cell pipeline'])
        self.plots_cells_after_sim.load(
            conf=plot_pipelines['cell cluster pipeline'])


    def unload(self) -> None:

        # Unload our superclass.
        super().unload()

        # Unload all subconfigurations of this configuration.
        self.plots_cell_after_sim.unload()
        self.plots_cells_after_sim.unload()

    # ..................{ PROPERTIES ~ after                }..................
    @property
    def is_after_sim(self) -> bool:
        return self.is_after_sim_save or self.is_after_sim_show


    @is_after_sim.setter
    @type_check
    def is_after_sim(self, is_after_sim: bool) -> None:
        self.is_after_sim_save = is_after_sim
        self.is_after_sim_show = is_after_sim

# ....................{ SUBCLASSES ~ item                 }....................
class SimConfExportPlotCell(SimConfExportABC):
    '''
    **Exported single-cell plot subconfiguration** (i.e., YAML-backed list item
    configuring the exportation of one or more images specific to a single cell
    from the simulation configuration file containing this item).
    '''

    # ..................{ SUPERCLASS                        }..................
    @classmethod
    @type_check
    def make_default(cls, yaml_list: YamlList) -> YamlListItemABC:

        # Duplicate the first plot in our default configuration file.
        return cls._make_loaded(conf={
            'name': yaml_list.get_item_name_uniquified(
                'Single Cell Plot ({})'),
            'type': 'voltage_membrane',
            'enabled': True,
        })


class SimConfExportPlotCells(
    SimConfVisualCellsYAMLMixin, SimConfExportABC):
    '''
    **Exported cell cluster plot subconfiguration** (i.e., YAML-backed list
    item configuring the exportation of one or more images applicable to all
    cells from the simulation configuration file containing this item).
    '''

    # ..................{ SUPERCLASS                        }..................
    @classmethod
    @type_check
    def make_default(cls, yaml_list: YamlList) -> YamlListItemABC:

        # Duplicate the first plot in our default configuration file.
        return cls._make_loaded(conf={
            'name': yaml_list.get_item_name_uniquified(
                'Cell Cluster Plot ({})'),
            'type': 'voltage_membrane',
            'enabled': True,
            'colorbar': {
                'autoscale': True,
                'minimum': -70.0,
                'maximum':  10.0,
            },
        })



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level Numpy-based array testing and validation facilities.
'''

# ....................{ IMPORTS                           }....................
import numpy as np
from betse.exceptions import BetseSimUnstableNaNException
from betse.util.type.types import type_check  #, ClassType, IterableTypes
from numpy import ndarray

# ....................{ EXCEPTIONS                        }....................
@type_check
def die_if_nan(array: ndarray) -> bool:
    '''
    Raise an exception if any element of any dimension of the passed Numpy
    array is a **NaN** (i.e., Not-a-Number).

    Parameters
    ----------
    array : ndarray
        Numpy array to be tested for NaN elements.

    Raises
    ----------

    See Also
    ----------
    :func:`is_nan`
        Further details.
    '''

    if is_nan(array):
        raise BetseSimUnstableNaNException(
            'Simulation instability detected '
            '(i.e., one or more NaNs in Numpy array). '
            'Consider a smaller time step, reduced gap junction radius, '
            'and/or reduced pump rate coefficients.'
        )

# ....................{ TESTERS                           }....................
@type_check
def is_nan(array: ndarray) -> bool:
    '''
    ``True`` only if any element of any dimension of the passed Numpy array is
    a **NaN** (i.e., Not-a-Number).

    Assuming the default Numpy configuration, Numpy produces NaN elements in
    all of the following numerical cases:

    * Division by zero (e.g., ``0/0``).
    * Division of infinity by infinity (e.g., ``inf/inf``).
    * Multplication of infinity by zero (e.g., ``0 * inf``, ``inf * 0``).
    * Subtraction of infinity by infinity (e.g., ``inf - inf``,
      ``(-inf) - (-inf)``).
    * Production of a complex number (e.g., ``np.sqrt(x)`` when ``x < 0``).
    * Performing the floating-point remainder of either:
      * A dividend that is infinite (e.g., ``np.fmod(x, y)`` when x is
        ``inf``).
      * A divisor that is zero (e.g., ``np.fmod(x, y)`` when y is 0).

    Caveats
    ----------
    **This high-level function should always be called in lieue of the
    low-level :func:`np.isnan` function,** which is space- and time-consumptive
    by compare to this function. Unlike this function, the :func:`np.isnan`
    returns a boolean array of the same shape as the passed array. In the
    common case of a two- or three-dimensional simulation matrix, passing this
    arrays to :func:`np.isnan` commonly produces a prohibitively large boolean
    array.

    This function circumvents such inefficiency by (in order):

    #. Reducing the passed array to a scalar value. If the local Numpy
       installation is linked against an optimized BLAS implementation (e.g.,
       ATLAS, OpenBLAS), this reduction is additionally parallelized across all
       available cores.
    #. Passing this scalar value to :func:`np.isnan`, producing an optimally
       space- and time-efficient scalar boolean value, which is returned as is.

    Parameters
    ----------
    array : ndarray
        Numpy array to be tested for NaN elements.

    Returns
    ----------
    bool
        ``True`` only if any element of any dimension of this array is a NaN.

    See Also
    ----------
    https://stackoverflow.com/a/6739580/2809027
        StackOverflow answer strongly inspiring this implementation.
    '''

    # Scalar value reduced from this array. Since Numpy propagates all NaNs in
    # mathematical operations, this value is guaranteed to be either:
    #
    # * If any element of any dimension of this array is a NaN, NaN.
    # * Else, an arbitrary non-NaN value.
    #
    # To parallelize this operation across all available cores under optimized
    # BLAS implementations, this scalar is produced with the BLAS-parallelized
    # dot product operator rather than the min() or sum() functions.
    array_scalar = np.dot(array, array)

    # Return true only if this scalar value is a NaN.
    return np.isnan(array_scalar)



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
YAML-backed simulation subconfiguration classes for exporting visuals (e.g.,
plots, animations).
'''

# ....................{ IMPORTS                           }....................
from abc import ABCMeta
from betse.lib.yaml.yamlalias import yaml_alias
# from betse.util.io.log import logs
from betse.util.type.decorator.deccls import abstractproperty
from betse.util.type.types import type_check, NumericSimpleTypes

# ....................{ SUPERCLASSES                      }....................
#FIXME: Non-ideal. Ideally, all networks subconfigurations should be refactored
#to leverage the YAML format specified by "SimConfVisualCellsYAMLMixin".
#FIXME: Rename to "SimConfColorbarABC" for clarity.
class SimConfVisualCellsABC(object, metaclass=ABCMeta):
    '''
    Abstract base class of all cell cluster visual subconfiguration subclasses,
    YAML-backed or otherwise.

    Attributes (Colorbar)
    ----------
    color_max : NumericSimpleTypes
        Maximum color value to be displayed by this visual's colorbar. Ignored
        if :attr:`is_color_autoscaled` is ``True``.
    color_min : NumericSimpleTypes
        Minimum color value to be displayed by this visual's colorbar. Ignored
        if :attr:`is_color_autoscaled` is ``True``.
    is_color_autoscaled : bool
        ``True`` if dynamically setting the minimum and maximum colorbar values
        for this visual to the minimum and maximum values flattened from
        the corresponding time series *or* ``False`` if statically setting
        these values to :attr:`color_min` and :attr:`color_max`.
    '''

    # ..................{ PROPERTIES                        }..................
    @abstractproperty
    def is_color_autoscaled(self) -> bool:
        pass

    @abstractproperty
    def color_min(self) -> NumericSimpleTypes:
        pass

    @abstractproperty
    def color_max(self) -> NumericSimpleTypes:
        pass


#FIXME: Rename to "SimConfColorbarMixin" for clarity.
class SimConfVisualCellsYAMLMixin(SimConfVisualCellsABC):
    '''
    Mixin of all **YAML-backed cell cluster visual subconfiguration** (i.e.,
    configuration of a single visual export parsed from the current
    YAML-formatted simulation configuration file) subclasses.
    '''

    # ..................{ ALIASES ~ colorbar                }..................
    is_color_autoscaled = yaml_alias("['colorbar']['autoscale']", bool)
    color_min = yaml_alias("['colorbar']['minimum']", NumericSimpleTypes)
    color_max = yaml_alias("['colorbar']['maximum']", NumericSimpleTypes)

# ....................{ SUBCLASSES                        }....................
#FIXME: Eliminate this subclass. For serializability, all configuration classes
#should be YAML-backed.
class SimConfVisualCellsNonYAML(SimConfVisualCellsABC):
    '''
    Cell cluster visual subconfiguration specific to the
    :mod:`betse.science.networks` package.

    Unlike all cell cluster visual subconfigurations, this class preserves
    backward compatibility by implementing superclass properties *without*
    calling the :func:`yaml_alias` function returning YAML-backed data
    descriptors. Subconfiguration changes are *not* propagated back to disk.
    '''

    # ..................{ INITIALIZERS                      }..................
    @type_check
    def __init__(
        self,
        is_color_autoscaled: bool,
        color_min: NumericSimpleTypes,
        color_max: NumericSimpleTypes,
    ) -> None:

        self._is_color_autoscaled = is_color_autoscaled
        self._color_min = color_min
        self._color_max = color_max

    # ..................{ PROPERTIES                        }..................
    @property
    def is_color_autoscaled(self) -> bool:
        return self._is_color_autoscaled

    @property
    def color_min(self) -> NumericSimpleTypes:
        return self._color_min

    @property
    def color_max(self) -> NumericSimpleTypes:
        return self._color_max



#!/usr/bin/env python3
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
YAML-backed simulation phase subconfigurations.
'''

# ....................{ IMPORTS                            }....................
from betse.lib.yaml.yamlalias import yaml_alias
from betse.lib.yaml.abc.yamlabc import YamlABC
# from betse.util.type.types import type_check, NumericSimpleTypes

# ....................{ SUBCLASSES                         }....................
#FIXME: Implement us up.
#FIXME: Refactor the Parameters.set_time_profile() method to leverage this.
class SimConfPhase(YamlABC):
    '''
    YAML-backed simulation phase subconfiguration, encapsulating the
    configuration of a single phase (e.g., seed, initialization, simulation)
    parsed from the current YAML-formatted simulation configuration file.

    Attributes (Path)
    ----------
    . : .
        .

    Attributes (Time)
    ----------
    . : .
        .
    '''

    # ..................{ ALIASES ~ wut                      }..................
    wut = yaml_alias("['wut']['wut']", int)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
**Autouse fixtures** (i.e., fixtures unconditionally applicable to all tests
within a given scope of the current test session).
'''

# ....................{ IMPORTS                            }....................
from betse.util.type.types import GeneratorType
from pytest import fixture

# ....................{ FIXTURES ~ package                 }....................
@fixture(scope='package', autouse=True)
def betse_init_package() -> None:
    '''
    **Autouse package initialization fixture** (i.e., fixture unconditionally
    applicable to all tests within the test package importing this fixture,
    which :mod:`pytest` automatically invokes once at session startup *after*
    collecting these tests but *before* running these tests).

    This fixture initializes the application metadata singleton -- a mandatory
    prerequisite before exercising *any* application functionality. Since
    functional (but *not* unit) tests already implicitly initialize this
    singleton, this fixture is typically only imported by the top-level unit
    testing subpackage (e.g., :mod:`betse_test.a00_unit`).

    See Also
    ----------
    :func:`_run_app_tests`
        Further details.
    '''

    # Make it so... or else.
    yield from _run_app_tests()

# ....................{ FIXTURES ~ session                }....................
#FIXME: This fixture is currently unused until we inevitably require this yet
#again. Ergo, this should be preserved for the sad future that comes.
# @fixture(scope='session', autouse=True)
# def betse_init_session() -> None:
#     '''
#     **Autouse session initialization fixture** (i.e., fixture unconditionally
#     applicable to all tests for the current test session, which :mod:`pytest`
#     automatically invokes once at session startup *after* collecting these
#     tests but *before* running these tests).
#
#     This fixture initializes the application metadata singleton -- a mandatory
#     prerequisite before exercising *any* application functionality. Since a
#     proper subset of testing fixtures and decorators (e.g., the
#     :func:`betse.util.test.pytest.mark.pytskip.skip_if_requirement` decorator)
#     assume this singleton to have already been initialized, this fixture is
#     typically imported by the top-level plugin (e.g.,
#     :mod:`betse_test.conftest`).
#     '''
#
#     pass

# ....................{ PRIVATE ~ runners                 }....................
def _run_app_tests() -> GeneratorType:
    '''
    Generator exercising all application-specific tests scoped by the calling
    fixture by ensuring that the application metadata singleton is initialized
    *before* the first such test and deinitialized *after* the last such test.

    Specifically, this generator (in order):

    #. Initializes the application metadata singleton.
    #. Yields control to the caller.
    #. Deinitializes this singleton, regardless of whether the caller raised an
       exception or not.

    See Also
    ----------
    :func:`init_app`
    :func:`deinit_app`
        Further details.
    '''

    # Initialize the application metadata singleton.
    init_app()

    # Yield control to the body of the caller's "with" block.
    try:
        yield
    # Deinitialize this singleton even if that block raised an exception.
    finally:
        deinit_app()

# ....................{ [DE]INITIALIZERS                  }....................
def init_app() -> None:
    '''
    Initialize the application metadata singleton (e.g., for unit tests, which
    fail to implicitly initialize this singleton ala functional tests).

    Specifically, this fixture (in order):

    #. Coerces the active Python interpreter into running **headless** (i.e.,
       with *no* access to a GUI display). Allowing headfull operation would
       would allow tests erroneously attempting to connect to an X11 server to
       locally succeed but remotely fail, as headless continuous integration
       (CI) pipelines typically have no access to an X11 server. Coercing
       headlessness ensures orthogonality between these cases by coercing the
       former to fail as well.
    #. Initializes the application core.
    #. Initializes all third-party dependencies thereof.
    #. Enables the default non-interactive matplotlib backend ``Agg``,
       *guaranteed* to be usable on all platforms. By default, matplotlib
       enables an interactive backend (e.g., ``Qt5Agg``) unsuitable for use
       under typically headless test automation.

    Motivation
    ----------
    This function performs early test-specific initialization of this
    application and dependencies thereof. Doing so prevents the magic
    :func:`betse.science.__init__` function from attempting to perform a
    subsequent test-agnostic initialization of either this application or
    dependencies on the first importation of the :mod:`betse.science`
    subpackage -- as in fixtures importing from that subpackage (e.g., the
    :mod:`betse_test._fixture.simconf.simconfer` fixture importing the
    :mod:`betse_test._fixture.simconf.simconfwrapper` submodule importing the
    :med:`betse.science.config.confwrap` submodule).
    '''

    # Defer heavyweight imports.
    from betse.util.app.meta import appmetaone
    from betse.util.os import displays
    from betse.util.test.pytest import pytests

    # If an application metadata singleton already exists, reduce to a noop.
    if appmetaone.is_app_meta():
        return
    # Else, no application metadata singleton currently exists.

    # Prepend a leading newline, which py.test curiously neglects to do itself.
    print('\n')

    # Print this initialization.
    pytests.output('Initializing application metadata singleton...')

    # Initialize a BETSE-specific application metadata singleton if the
    # appmetaone.set_app_meta() function has yet to be called.
    app_meta = appmetaone.set_app_meta_betse_if_unset()

    # Coerce the active Python interpreter into running headless *AFTER*
    # initializing this singleton, which enables the default logging
    # configuration to which this setter logs this operation.
    #
    # Note that this operation technically needs to be performed:
    #
    # * Only once for the entire test suite when py.test is *NOT* parallelized
    #   with "xdist", in which case all tests run in the same process and hence
    #   share the same global variables.
    # * Once for each test when py.test is parallelized with "xdist", in which
    #   case each test is run in a distinct subprocess and hence does *NOT*
    #   share the same global variables.
    #
    # Since setting global variables is fast, doing so here transparently
    # supports both use cases detailed above with no discernable downside. See
    # the docstring for additional commentary.
    displays.set_headless(True)

    # Initialize all mandatory third-party dependencies with a standard
    # non-interactive matplotlib backend guaranteed to exist *AFTER* coercing
    # the active Python interpreter into running headless. Why? Because
    # dependencies typically detect headless environments.
    app_meta.init_libs(matplotlib_backend_name='Agg')

    # Print the completion of this initialization.
    pytests.output('Initialized application metadata singleton.')


def deinit_app() -> None:
    '''
    Deinitialize the application metadata singleton.
    '''

    # Defer heavyweight imports.
    from betse.util.app.meta import appmetaone
    from betse.util.test.pytest import pytests

    # If no application metadata singleton currently exists, reduce to a noop.
    if not appmetaone.is_app_meta():
        return
    # Else, an application metadata singleton currently exists.

    # Prepend a leading newline, which py.test curiously neglects to do itself.
    print('\n')

    # Print this deinitialization.
    pytests.output('Deinitializing application metadata singleton...')

    # Deinitialize the application metadata singleton.
    appmetaone.deinit()

    # Print this deinitialization.
    pytests.output('Deinitialized application metadata singleton.')



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Project-wide **mathematical interpolators** (i.e., low-level callables
interpolating two- and three-dimensional meshes defined from one set of points
onto another set of points).
'''

# ....................{ IMPORTS                            }....................
from beartype.typing import (
    Callable,
)
from betse.exceptions import BetseMathInterpolationException
from betse.util.type.typehints import NDArrayNdim1, NDArrayNdim2
# from functools import partial
from numpy import ndarray
from scipy.interpolate import (
    RectBivariateSpline,
    # bisplev,
    # bisplrep,
)

# ....................{ INTERPOLATORS                      }....................
def interp2d_linear(
    x: NDArrayNdim1,
    y: NDArrayNdim1,
    z: NDArrayNdim2,
) -> (
    Callable[[ndarray, ndarray], ndarray]):
    '''
    Create and return a new **two-dimensional interpolator** (i.e., low-level
    callable interpolating a source mesh defined on the passed set of X and Y
    coordinates comprising a so-called "regular grid" and Z values (e.g.,
    colours) onto a target mesh defined on a new set of X and Y coordinates
    passed to that callable, which then returns the new Z values produced by
    that interpolation).

    This interpolator implements the "regular grid" code path of the obsolete
    :class:`scipy.interpolate.interp2d` class (passed the parameters
    ``kind='linear'`` and ``fill_value=0.0``) in terms of non-obsolete
    functionality defined by the :mod:`scipy.interpolate` subpackage, improving
    backward compatibility with existing code expecting that class to exist.

    Caveats
    -------
    This interpolator requires that the passed arrays satisfy these constraints:

    .. code-block:: python

       # The X and Y coordinates are *ALL* 1-dimensional: i.e.,
       x.ndim == y.ndim == 1

       # The Z values are 2-dimensional: i.e.,
       z.ndim == 2

       # The total number of X and Y coordinates differ: i.e.,
       x.size != y.size

       # The total number of Z values is the product of:
       # * The total number of X coordinates and...
       # * The total number of Y coordinates.
       z.size == x.size * y.size

    If any one of these constraints is *not* the case, an exception is raised.

    Parameters
    ----------
    x: NDArrayNdim1
        Source 1-dimensional X coordinates to interpolate from.
    y: NDArrayNdim1
        Source 1-dimensional Y coordinates to interpolate from.
    z: NDArrayNdim2
        Source 2-dimensional Z values to interpolate from.

    Returns
    -------
    Callable[[ndarray, ndarray], ndarray]
        Two-dimensional interpolator as described above. The signature of this
        callable should be assumed to be:

        .. code-block:: python

           def interpolator(x_new: ndarray, y_new: ndarray) -> ndarray: ...

    Raises
    ------
    BetseMathInterpolationException
        If any of the above constraints are violated.
    '''

    # If the total number of Z values is *NOT* the product of the total number
    # of X and Y coordinates, raise an exception.
    if z.size != x.size * y.size:
        raise BetseMathInterpolationException(
            f'Interpolation Z shape not product of X and Y coordinates shape '
            f'(i.e., "z.size = {z.size}" != '
            f'"x.size = {x.size}" * "y.size = {y.size}").'
        )
    # Else, the total number of Z values is the product of the total number
    # of X and Y coordinates.

    #FIXME: This doesn't appear to hold. Square shapes are fine. *shrug*
    # # If the number of X and Y coordinates are the same, raise an exception.
    # if x.size == y.size:
    #     raise BetseMathInterpolationException(
    #         f'Interpolation X and Y coordinates share same size (i.e., '
    #         f'"x.size = {x.size}" and "y.size = {y.size}" equal).'
    #     )
    # # Else, the number of X and Y coordinates differ.

    # #FIXME: Comment this out to test our new interp2d() replacement, which is
    # #currently broken for unknown reasons. For example, this test fails:
    # #    ./pytest -k test_cli_sim_compat
    # from scipy.interpolate import interp2d
    # return interp2d(x, y, z, kind='linear', fill_value=0.0)

    #FIXME; Non-ideal. Modern SciPy documentation strongly advises usage of:
    #    For legacy code, nearly bug-for-bug compatible replacements are
    #    `RectBivariateSpline` on regular grids, and `bisplrep`/`bisplev` for
    #    scattered 2D data.
    #    
    #    In new code, for regular grids use `RegularGridInterpolator` instead.
    #    For scattered data, prefer `LinearNDInterpolator` or
    #    `CloughTocher2DInterpolator`.
    #
    #    For more details see
    #    https://scipy.github.io/devdocs/tutorial/interpolate/interp_transition_guide.html
    #
    #We're fairly certain we should be leveraging "RectBivariateSpline" rather
    #than "bisplrep" here. The above SciPy documentation as well as Ally's
    #perfect memory suggests a similar finding. Consider refactoring, please.
    #
    #In short, there appear to exist two code paths. The first code path is
    #given by:
    #    # If the total number of X and Y coordinates differ *AND*...
    #    if x.size != y.size and z.size == len(x) * len(y):
    #       then do "RectBivariateSpline" stuff.
    #    elif len(x) == len(y) == len(z):
    #       then do "bisplrep" stuff.
    #    else:
    #       raise Exception(!)

    # Linear bivariate spline approximation over this rectangular mesh.
    #
    # Note that:
    # * The "kx=1" and "ky=1" parameters produce a linear interpolation.
    # * The "z.T" transpose reproduces the behaviour of the obsolete
    #   scipy.interpolate.interp2d() function that this higher-level function
    #   descended from and preserves backward API compatibility with.
    bivariate_spline = RectBivariateSpline(x, y, z.T, kx=1, ky=1)

    # Linear bivariate spline interpolator over this rectangular mesh.
    #
    # Note that:
    # * The ".T" transpose reproduces the behaviour of the obsolete
    #   scipy.interpolate.interp2d() function that this higher-level function
    #   descended from and preserves backward API compatibility with.
    bivariate_spline_interpolator = lambda x_new, y_new: (
        bivariate_spline(x_new, y_new).T)

    # Return this interpolator.
    return bivariate_spline_interpolator

    #FIXME: Currently unused, but preserved for posterity. *shrug*
    # # Linear B-spline representation of this two-dimensional surface.
    # #
    # # Note that the "kx=1" and "ky=1" parameters effectively produce a linear
    # # interpolation.
    # tck = bisplrep(x, y, z, kx=1, ky=1)
    #
    # # Return the desired interpolator for this linear B-spline representation.
    # #
    # # Note that usage of partial() here is required to avoid infinite recursion
    # # from the standard "pickle" and third-party "dill" modules. In particular,
    # # attempting to replace this partial() approach with a simpler lambda
    # # function or closure induces infinite recursion for unknown reasons.
    # return partial(_interpolator2d_linear, _tck=tck)

# ....................{ PRIVATE ~ interpolators            }....................
#FIXME: Currently unused, but preserved for posterity. *shrug*
# #FIXME: This function is intentionally *NOT* type-checked via @beartype, as both
# #the parameters and return values appear to unexpectedly expect NumPy scalars as
# #well as arrays. For now, let's just permissively run with duck typing. *sigh*
# def _interpolator2d_linear_scattered(x_new, y_new, _tck) -> ndarray:
#     '''
#     Z values (e.g., colours) produced by interpolating from the passed source
#     linear B-spline representation describing the source mesh defined on the
#     set of X and Y coordinates and Z values (e.g., colours) previously passed to
#     the :func:`.interp2d_linear` function onto a target mesh defined on the
#     passed X and Y coordinates.
#
#     Parameters
#     ----------
#     x: ndarray
#         Target X coordinates to interpolate onto.
#     y: ndarray
#         Target Y coordinates to interpolate onto.
#     _tck : ndarray
#         List ``[tx, ty, c, kx, ky]`` containing the knots ``(tx, ty)`` and
#         coefficients ``(c)`` of the bivariate B-spline representation of the
#         surface along with the degree of the spline, typically obtained by a
#         prior call to the :func:`.bisplrep` function.
#
#     Returns
#     -------
#     ndarray
#         Z values (e.g., colours) produced by this interpolation.
#     '''
#
#     # Perform this interpolation. Note that the transpose (i.e., ".T") mimics
#     # the behaviour of the original scipy.interpolate.interp2d() function.
#     return bisplev(x_new, y_new, _tck).T



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
**Export-specific enumeration types** (i.e., :class:`enum.Enum` subclasses
expressing mutually exclusive choices specific to post-simulation exports).
'''

# ....................{ IMPORTS                           }....................
from betse.util.type.enums import make_enum

# ....................{ ENUMS                             }....................
SimExportType = make_enum(
    class_name='SimExportType',
    member_names=('ANIM', 'CSV', 'PLOT',),
    doc='''
    Enumeration of all supported types of **post-simulation exports** (e.g.,
    plots, animations, comma-separated value (CSV) files).

    Attributes
    ----------
    ANIM : enum
        Post-simulation animations, visualizing the cell cluster across one or
        more sampled time steps of an initialization or simulation phase as the
        same number of frames of an animated video.
    CSV : enum
        Post-simulation comma-separated value (CSV) files, aggregating raw data
        for the cell cluster across one or more sampled time steps of an
        initialization or simulation phase.
    PLOT : enum
        Post-simulation plots, visualizing the cell cluster across one or more
        sampled time steps of an initialization or simulation phase as a single
        non-animated image.
    ''',
)



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

# ....................{ IMPORTS                           }....................
import numpy as np
from betse.science import sim_toolbox as stb
# from scipy.ndimage import gaussian_filter
# from betse.science.math import finitediff as fd

# ....................{ GETTERS                           }....................
def getFlow(sim, cells, p):
    '''
    Calculate electroosmotic fluid flow in both cells and extracellular
    networks using the well-known Stokes-Equation.
    '''

    # First do extracellular space electroosmotic flow.
    if p.is_ecm:
        # Use Helmholtz-Smoluchowski equation.
        muFx = (
            -p.eo*p.er*(1 / p.mu_water) *
            sim.E_env_x *
            sim.v_env.reshape(cells.X.shape) *
            sim.D_env_weight
        )
        muFy = (
            -p.eo*p.er*(1 / p.mu_water) *
            sim.E_env_y *
            #FIXME: To support non-square grids, this should probably be:
            #    sim.v_env.reshape(cells.Y.shape) *
            #Since we only support square grids currently, it's fine for now.
            sim.v_env.reshape(cells.X.shape) *
            sim.D_env_weight
        )

        uxo = np.dot(cells.lapENVinv, -muFx.ravel())
        uyo = np.dot(cells.lapENVinv, -muFy.ravel())

        _, sim.u_env_x, sim.u_env_y, _, _, _ = stb.HH_Decomp(uxo, uyo, cells)

    # -------Next do flow through gap junction connected cells-----------------

    # net force is the electrostatic body force on net volume charge in cells:
    Fxc = sim.E_cell_x*sim.rho_cells*(1/p.mu_water)*p.gj_surface
    Fyc = sim.E_cell_y*sim.rho_cells*(1/p.mu_water)*p.gj_surface

    # Calculate flow under body forces using Stokes flow:
    u_gj_xo = np.dot(cells.lapGJinv, -Fxc)
    u_gj_yo = np.dot(cells.lapGJinv, -Fyc)

    # Coerce the flow to be divergence-free via the standard Helmholtz-Hodge
    # decomposition method.
    _, sim.u_cells_x, sim.u_cells_y, _, _, _ = cells.HH_cells(
        u_gj_xo, u_gj_yo, rot_only=True)

    #---Wastelands---------------------------------------------------------------------------------------------------
    # --Alternative method #1 calculates fluid flow in terms of slip velocity only:----------------------------------

    # scaleF =  (sim.ko_env*p.cell_height) # conversion concentrating charge density in the double layer
    #
    # # total charge in an env-grid square in Coulombs:
    # q_env = sim.rho_env.reshape(cells.X.shape)*(cells.delta**2)*(p.cell_height)
    #
    # # slip velocity forces:
    # # This assumes that the *screening* charge is pulled on by the self-generated electric field.
    # # Further investigation may be warranted in the future as to the sign of the slip velocity
    # muFx = -(1/p.mu_water)*sim.E_env_x*q_env*scaleF*sim.D_env_weight
    # muFy = -(1/p.mu_water)*sim.E_env_y*q_env*scaleF*sim.D_env_weight
    #
    # _, sim.u_env_x, sim.u_env_y, _, _, _ = stb.HH_Decomp(muFx, muFy, cells)


    # --Alternative method #2 calculates fluid flow in terms of slip velocity only using Helmholtz-Smoluchowski:-----

    # slip velocity forces:
    # This assumes that the *screening* charge is pulled on by the self-generated electric field.
    # Further investigation may be warranted in the future as to the sign of the slip velocity

    # vme = np.zeros(sim.edl)
    # vme[cells.map_mem2ecm] = sim.vm/2
    # vme = vme.reshape(cells.X.shape)
    #
    # muFx = (1/p.mu_water)*sim.E_env_x*(vme)*p.eo*p.er
    # muFy = (1/p.mu_water)*sim.E_env_y*(vme)*p.eo*p.er
    #
    # _, sim.u_env_x, sim.u_env_y, _, _, _ = stb.HH_Decomp(muFx, muFy, cells)

    # Alternative method #3 calculates flow in terms of curl component of current density:--------------------------
    # cc = sim.cc_env.mean(axis=0).reshape(cells.X.shape)
    # zz = sim.zs.mean()
    #
    # sim.u_env_x = -sim.J_env_x / (p.F * cc * zz)
    # sim.u_env_y = -sim.J_env_y / (p.F * cc * zz)

